# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import uuid
from collections import defaultdict
from collections.abc import Callable
from concurrent.futures import Future
from contextlib import suppress
from typing import TYPE_CHECKING
from unittest import mock
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest
from openlineage.client import OpenLineageClient
from openlineage.client.transport import ConsoleTransport
from openlineage.client.transport.console import ConsoleConfig
from uuid6 import uuid7

from airflow.models import DAG, DagRun, TaskInstance

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import BaseOperator
else:
    from airflow.models.baseoperator import BaseOperator  # type: ignore[no-redef]

from airflow.providers.openlineage.extractors.base import OperatorLineage
from airflow.providers.openlineage.plugins.adapter import OpenLineageAdapter
from airflow.providers.openlineage.plugins.listener import OpenLineageListener
from airflow.providers.openlineage.utils.selective_enable import disable_lineage, enable_lineage
from airflow.utils import timezone, types
from airflow.utils.state import DagRunState, State

from tests_common.test_utils.compat import EmptyOperator, PythonOperator
from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.db import clear_db_runs
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

EXPECTED_TRY_NUMBER_1 = 1

TRY_NUMBER_BEFORE_EXECUTION = 0
TRY_NUMBER_RUNNING = 0
TRY_NUMBER_FAILED = 0
TRY_NUMBER_SUCCESS = 0
TRY_NUMBER_AFTER_EXECUTION = 0

if TYPE_CHECKING:
    from airflow.sdk.execution_time.task_runner import RuntimeTaskInstance


class TemplateOperator(BaseOperator):
    template_fields = ["df"]

    def __init__(self, df, *args, **kwargs):
        self.df = df
        super().__init__(*args, **kwargs)

    def execute(self, context):
        return self.df


def render_df():
    return pd.DataFrame({"col": [1, 2]})


def regular_call(self, callable, callable_name, use_fork):
    callable()


class MockExecutor:
    def __init__(self, *args, **kwargs):
        self.submitted = False
        self.succeeded = False
        self.result = None

    def submit(self, fn, /, *args, **kwargs):
        self.submitted = True
        try:
            fn(*args, **kwargs)
            self.succeeded = True
        except Exception:
            pass
        return MagicMock()

    def shutdown(self, *args, **kwargs):
        print("Shutting down")


@pytest.mark.skipif(AIRFLOW_V_3_0_PLUS, reason="Airflow 2 tests")
class TestOpenLineageListenerAirflow2:
    @patch("airflow.models.TaskInstance.xcom_push")
    @patch("airflow.models.BaseOperator.render_template")
    def test_listener_does_not_change_task_instance(self, render_mock, xcom_push_mock):
        render_mock.return_value = render_df()
        date = timezone.datetime(2022, 1, 1)
        dag = DAG(
            "test",
            schedule=None,
            start_date=date,
            user_defined_macros={"render_df": render_df},
            params={"df": {"col": [1, 2]}},
        )
        t = TemplateOperator(task_id="template_op", dag=dag, do_xcom_push=True, df=dag.param("df"))
        run_id = str(uuid.uuid1())
        dag.create_dagrun(
            run_id=run_id,
            data_interval=(date, date),
            run_type=types.DagRunType.MANUAL,
            state=DagRunState.RUNNING,
        )
        ti = TaskInstance(t, run_id=run_id)
        ti.check_and_change_state_before_execution()  # make listener hook on running event
        ti._run_raw_task()

        # check if task returns the same DataFrame
        pd.testing.assert_frame_equal(xcom_push_mock.call_args.kwargs["value"], render_df())

        # check if render_template method always get the same unrendered field
        assert not isinstance(render_mock.call_args.args[0], pd.DataFrame)

    def _setup_mock_listener(self, mock_listener: mock.Mock, captured_try_numbers: dict[str, int]) -> None:
        """Sets up the mock listener with side effects to capture try numbers for different task instance events.

        :param mock_listener: The mock object for the listener manager.
        :param captured_try_numbers: A dictionary to store captured try numbers keyed by event names.

        This function iterates through specified event names and sets a side effect on the corresponding
        method of the listener manager's hook. The side effect is a nested function that captures the try number
        of the task instance when the method is called.

        :Example:

            captured_try_numbers = {}
            mock_listener = Mock()
            _setup_mock_listener(mock_listener, captured_try_numbers)
            # After running a task, captured_try_numbers will have the try number captured at the moment of
            execution for specified methods. F.e. {"running": 1, "success": 2} for on_task_instance_running and
            on_task_instance_success methods.
        """

        def capture_try_number(method_name):
            def inner(*args, **kwargs):
                captured_try_numbers[method_name] = kwargs["task_instance"].try_number

            return inner

        for event in ["running", "success", "failed"]:
            getattr(
                mock_listener.return_value.hook, f"on_task_instance_{event}"
            ).side_effect = capture_try_number(event)

    def _create_test_dag_and_task(
        self, python_callable: Callable, scenario_name: str
    ) -> tuple[DagRun, TaskInstance]:
        """Creates a test DAG and a task for a custom test scenario.

        :param python_callable: The Python callable to be executed by the PythonOperator.
        :param scenario_name: The name of the test scenario, used to uniquely name the DAG and task.

        :return: TaskInstance: The created TaskInstance object.

        This function creates a DAG and a PythonOperator task with the provided
        python_callable. It generates a unique run ID and creates a DAG run. This
        setup is useful for testing different scenarios in Airflow tasks.

            :Example:

                def sample_callable(**kwargs):
                    print("Hello World")

            task_instance = _create_test_dag_and_task(sample_callable, "sample_scenario")
            # Use task_instance to simulate running a task in a test.
        """
        date = timezone.datetime(2022, 1, 1)
        dag = DAG(
            f"test_{scenario_name}",
            schedule=None,
            start_date=date,
        )
        t = PythonOperator(task_id=f"test_task_{scenario_name}", dag=dag, python_callable=python_callable)
        run_id = str(uuid.uuid1())
        dagrun = dag.create_dagrun(
            run_id=run_id,
            data_interval=(date, date),
            run_type=types.DagRunType.MANUAL,
            state=DagRunState.RUNNING,
            execution_date=date,  # type: ignore
        )
        if AIRFLOW_V_3_0_PLUS:
            task_instance = TaskInstance(t, run_id=run_id, dag_version_id=dagrun.created_dag_version_id)
        else:
            task_instance = TaskInstance(t, run_id=run_id)  # type: ignore
        return dagrun, task_instance

    def _create_listener_and_task_instance(self) -> tuple[OpenLineageListener, TaskInstance]:
        """Creates and configures an OpenLineageListener instance and a mock TaskInstance for testing.

        :return: A tuple containing the configured OpenLineageListener and TaskInstance.

        This function instantiates an OpenLineageListener, sets up its required properties with mock objects, and
        creates a mock TaskInstance with predefined attributes. This setup is commonly used for testing the
        interaction between an OpenLineageListener and a TaskInstance in Airflow.

        :Example:

            listener, task_instance = _create_listener_and_task_instance()
            # Now you can use listener and task_instance in your tests to simulate their interaction.
        """

        def mock_dag_id(dag_id, logical_date, clear_number):
            return f"{logical_date.isoformat()}.{dag_id}.{clear_number}"

        def mock_task_id(dag_id, task_id, try_number, logical_date, map_index):
            return f"{logical_date.isoformat()}.{dag_id}.{task_id}.{try_number}.{map_index}"

        listener = OpenLineageListener()
        listener.extractor_manager = mock.Mock()

        metadata = mock.Mock()
        metadata.run_facets = {"run_facet": 1}
        listener.extractor_manager.extract_metadata.return_value = metadata

        adapter = mock.Mock()
        adapter.build_dag_run_id.side_effect = mock_dag_id
        adapter.build_task_instance_run_id.side_effect = mock_task_id
        adapter.start_task = mock.Mock()
        adapter.fail_task = mock.Mock()
        adapter.complete_task = mock.Mock()
        listener.adapter = adapter
        if AIRFLOW_V_3_0_PLUS:
            task_instance = TaskInstance(task=mock.Mock(), dag_version_id=mock.MagicMock())
        else:
            task_instance = TaskInstance(task=mock.Mock())  # type: ignore
        task_instance.dag_run = DagRun()
        task_instance.dag_run.run_id = "dag_run_run_id"
        task_instance.dag_run.data_interval_start = None
        task_instance.dag_run.data_interval_end = None
        task_instance.dag_run.clear_number = 0
        task_instance.dag_run.execution_date = timezone.datetime(2020, 1, 1, 1, 1, 1)
        task_instance.task = mock.Mock()
        task_instance.task.task_id = "task_id_from_task_and_not_ti"
        task_instance.task.dag = mock.Mock()
        task_instance.task.dag.dag_id = "dag_id"
        task_instance.task.dag.description = "Test DAG Description"
        task_instance.task.dag.doc_md = None
        task_instance.task.dag.owner = "Test Owner"
        task_instance.task.dag.tags = ["tag1", "tag2"]
        task_instance.task.owner = "task_owner"
        task_instance.task.doc = None
        task_instance.task.doc_md = "Test TASK Description"
        task_instance.task.doc_json = None
        task_instance.task.doc_yaml = None
        task_instance.task.doc_rst = None
        task_instance.task.inlets = []
        task_instance.task.outlets = []
        task_instance.dag_id = "dag_id"
        task_instance.run_id = "dag_run_run_id"
        task_instance.task_id = "task_id"
        task_instance.try_number = 1
        task_instance.state = State.RUNNING
        task_instance.start_date = timezone.datetime(2023, 1, 1, 13, 1, 1)
        task_instance.end_date = timezone.datetime(2023, 1, 3, 13, 1, 1)
        task_instance.logical_date = timezone.datetime(2020, 1, 1, 1, 1, 1)
        task_instance.map_index = -1
        task_instance.next_method = None  # Ensure this is None to reach start_task
        task_instance.get_template_context = mock.MagicMock()  # type: ignore[method-assign]
        task_instance.get_template_context.return_value = defaultdict(mock.MagicMock)
        task_instance.get_template_context()["task_reschedule_count"] = 0

        return listener, task_instance

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_mapped_task_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_start_task_is_called_with_proper_arguments(
        self,
        mock_get_airflow_mapped_task_facet,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
    ):
        """Tests that the 'start_task' method of the OpenLineageAdapter is invoked with the correct arguments.

        The test checks that the job name, job description, event time, and other related data are
        correctly passed to the adapter. It also verifies that custom facets and Airflow run facets are
        correctly retrieved and included in the call. This ensures that all relevant data, including custom
        and Airflow-specific metadata, is accurately conveyed to the adapter during the initialization of a task,
        reflecting the comprehensive tracking of task execution contexts."""

        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_airflow_mapped_task_facet.return_value = {"mapped_facet": 1}
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        listener.on_task_instance_running(None, task_instance, None)
        listener.adapter.start_task.assert_called_once_with(
            run_id="2020-01-01T01:01:01+00:00.dag_id.task_id.1.-1",
            job_name="dag_id.task_id",
            job_description="Test TASK Description",
            job_description_type="text/markdown",
            event_time="2023-01-01T13:01:01+00:00",
            nominal_start_time=None,
            nominal_end_time=None,
            owners=["task_owner"],
            tags=["tag1", "tag2"],
            task=listener.extractor_manager.extract_metadata(),
            run_facets={
                "parent": 4,
                "mapped_facet": 1,
                "custom_user_facet": 2,
                "airflow_run_facet": 3,
                "debug": "packages",
            },
        )

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_mapped_task_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_start_task_is_called_with_dag_owners_when_task_owner_is_default(
        self,
        mock_get_airflow_mapped_task_facet,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_airflow_mapped_task_facet.return_value = {"mapped_facet": 1}
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        task_instance.task.owner = "airflow"  # Simulate default owner on task to force fallback to DAG owner
        listener.on_task_instance_running(None, task_instance, None)
        assert listener.adapter.start_task.call_args.kwargs["owners"] == ["Test Owner"]

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_mapped_task_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_job_name")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_start_task_is_called_with_dag_description_when_task_doc_is_empty(
        self,
        mock_get_job_name,
        mock_get_airflow_mapped_task_facet,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_job_name.return_value = "job_name"
        mock_get_airflow_mapped_task_facet.return_value = {"mapped_facet": 1}
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        task_instance.task.doc_md = None  # Simulate lack of task doc to force fallback to DAG description
        listener.on_task_instance_running(None, task_instance, None)
        assert listener.adapter.start_task.call_args.kwargs["job_description"] == "Test DAG Description"
        assert listener.adapter.start_task.call_args.kwargs["job_description_type"] == "text/plain"

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_fail_task_is_called_with_proper_arguments(
        self,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
        time_machine,
    ):
        """Tests that the 'fail_task' method of the OpenLineageAdapter is invoked with the correct arguments.

        This test ensures that the job name is accurately retrieved and included, along with the generated
        run_id and task metadata. By mocking the job name retrieval and the run_id generation,
        the test verifies the integrity and consistency of the data passed to the adapter during task
        failure events, thus confirming that the adapter's failure handling is functioning as expected.
        """
        time_machine.move_to(timezone.datetime(2023, 1, 3, 13, 1, 1))

        listener, task_instance = self._create_listener_and_task_instance()
        task_instance.logical_date = timezone.datetime(2020, 1, 1, 1, 1, 1)
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow": {"task": "..."}}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        err = ValueError("test")
        listener.on_task_instance_failed(
            previous_state=None, task_instance=task_instance, error=err, session=None
        )
        listener.adapter.fail_task.assert_called_once_with(
            end_time="2023-01-03T13:01:01+00:00",
            job_name="dag_id.task_id",
            run_id="2020-01-01T01:01:01+00:00.dag_id.task_id.1.-1",
            owners=["task_owner"],
            tags=["tag1", "tag2"],
            nominal_start_time=None,
            nominal_end_time=None,
            job_description="Test TASK Description",
            job_description_type="text/markdown",
            task=listener.extractor_manager.extract_metadata(),
            run_facets={
                "parent": 4,
                "custom_user_facet": 2,
                "airflow": {"task": "..."},
                "debug": "packages",
            },
            error=err,
        )

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_mapped_task_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_fail_task_is_called_with_dag_owners_when_task_owner_is_default(
        self,
        mock_get_airflow_mapped_task_facet,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_airflow_mapped_task_facet.return_value = {"mapped_facet": 1}
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        task_instance.task.owner = "airflow"  # Simulate default owner on task to force fallback to DAG owner
        err = ValueError("test")
        listener.on_task_instance_failed(
            previous_state=None, task_instance=task_instance, error=err, session=None
        )
        assert listener.adapter.fail_task.call_args.kwargs["owners"] == ["Test Owner"]

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_mapped_task_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_fail_task_is_called_with_dag_description_when_task_doc_is_empty(
        self,
        mock_get_airflow_mapped_task_facet,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_airflow_mapped_task_facet.return_value = {"mapped_facet": 1}
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        task_instance.task.doc_md = None  # Simulate lack of task doc to force fallback to DAG description
        err = ValueError("test")
        listener.on_task_instance_failed(
            previous_state=None, task_instance=task_instance, error=err, session=None
        )
        assert listener.adapter.fail_task.call_args.kwargs["job_description"] == "Test DAG Description"
        assert listener.adapter.fail_task.call_args.kwargs["job_description_type"] == "text/plain"

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_complete_task_is_called_with_proper_arguments(
        self,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
        time_machine,
    ):
        """Tests that the 'complete_task' method of the OpenLineageAdapter is called with the correct arguments.

        It checks that the job name is correctly retrieved and passed,
        along with the run_id and task metadata. The test also simulates changes in the try_number
        attribute of the task instance, as it would occur in Airflow, to ensure that the run_id is updated
        accordingly. This helps confirm the consistency and correctness of the data passed to the adapter
        during the task's lifecycle events.
        """
        time_machine.move_to(timezone.datetime(2023, 1, 3, 13, 1, 1))

        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow": {"task": "..."}}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        listener.on_task_instance_success(None, task_instance, None)
        # This run_id will be different as we did NOT simulate increase of the try_number attribute,
        # which happens in Airflow < 2.10.
        calls = listener.adapter.complete_task.call_args_list
        assert len(calls) == 1
        assert calls[0][1] == dict(
            end_time="2023-01-03T13:01:01+00:00",
            job_name="dag_id.task_id",
            run_id=f"2020-01-01T01:01:01+00:00.dag_id.task_id.{EXPECTED_TRY_NUMBER_1}.-1",
            task=listener.extractor_manager.extract_metadata(),
            owners=["task_owner"],
            tags=["tag1", "tag2"],
            nominal_start_time=None,
            nominal_end_time=None,
            job_description="Test TASK Description",
            job_description_type="text/markdown",
            run_facets={
                "parent": 4,
                "custom_user_facet": 2,
                "airflow": {"task": "..."},
                "debug": "packages",
            },
        )

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_mapped_task_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_complete_task_is_called_with_dag_owners_when_task_owner_is_default(
        self,
        mock_get_airflow_mapped_task_facet,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_airflow_mapped_task_facet.return_value = {"mapped_facet": 1}
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        task_instance.task.owner = "airflow"  # Simulate default owner on task to force fallback to DAG owner
        listener.on_task_instance_success(None, task_instance, None)
        assert listener.adapter.complete_task.call_args.kwargs["owners"] == ["Test Owner"]

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_mapped_task_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_complete_task_is_called_with_dag_description_when_task_doc_is_empty(
        self,
        mock_get_airflow_mapped_task_facet,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_airflow_mapped_task_facet.return_value = {"mapped_facet": 1}
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        task_instance.task.doc_md = None  # Simulate lack of task doc to force fallback to DAG description
        listener.on_task_instance_success(None, task_instance, None)
        assert listener.adapter.complete_task.call_args.kwargs["job_description"] == "Test DAG Description"
        assert listener.adapter.complete_task.call_args.kwargs["job_description_type"] == "text/plain"

    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_on_task_instance_running_correctly_calls_openlineage_adapter_run_id_method(self):
        """Tests the OpenLineageListener's response when a task instance is in the running state.

        This test ensures that when an Airflow task instance transitions to the running state,
        the OpenLineageAdapter's `build_task_instance_run_id` method is called exactly once with the correct
        parameters derived from the task instance.
        """
        listener, task_instance = self._create_listener_and_task_instance()
        listener.on_task_instance_running(None, task_instance, None)
        listener.adapter.build_task_instance_run_id.assert_called_once_with(
            dag_id="dag_id",
            task_id="task_id",
            logical_date=timezone.datetime(2020, 1, 1, 1, 1, 1),
            try_number=1,
            map_index=-1,
        )

    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_on_task_instance_failed_correctly_calls_openlineage_adapter_run_id_method(self):
        """Tests the OpenLineageListener's response when a task instance is in the failed state.

        This test ensures that when an Airflow task instance transitions to the failed state,
        the OpenLineageAdapter's `build_task_instance_run_id` method is called exactly once with the correct
        parameters derived from the task instance.
        """
        listener, task_instance = self._create_listener_and_task_instance()
        on_task_failed_kwargs = {"error": ValueError("test")}

        listener.on_task_instance_failed(
            previous_state=None, task_instance=task_instance, **on_task_failed_kwargs, session=None
        )
        listener.adapter.build_task_instance_run_id.assert_called_once_with(
            dag_id="dag_id",
            task_id="task_id",
            logical_date=timezone.datetime(2020, 1, 1, 1, 1, 1),
            try_number=1,
            map_index=-1,
        )

    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_on_task_instance_success_correctly_calls_openlineage_adapter_run_id_method(self):
        """Tests the OpenLineageListener's response when a task instance is in the success state.

        This test ensures that when an Airflow task instance transitions to the success state,
        the OpenLineageAdapter's `build_task_instance_run_id` method is called exactly once with the correct
        parameters derived from the task instance.
        """
        listener, task_instance = self._create_listener_and_task_instance()
        listener.on_task_instance_success(None, task_instance, None)
        listener.adapter.build_task_instance_run_id.assert_called_once_with(
            dag_id="dag_id",
            task_id="task_id",
            logical_date=timezone.datetime(2020, 1, 1, 1, 1, 1),
            try_number=EXPECTED_TRY_NUMBER_1,
            map_index=-1,
        )

    @mock.patch("airflow.models.taskinstance.get_listener_manager")
    def test_listener_on_task_instance_failed_is_called_before_try_number_increment(self, mock_listener):
        """Validates the listener's on-failure method is called before try_number increment happens.

        This test ensures that when a task instance fails, Airflow's listener method for
        task failure (`on_task_instance_failed`) is invoked before the increment of the
        `try_number` attribute happens. A custom exception simulates task failure, and the test
        captures the `try_number` at the moment of this method call.
        """
        captured_try_numbers = {}
        self._setup_mock_listener(mock_listener, captured_try_numbers)

        # Just to make sure no error interferes with the test, and we do not suppress it by accident
        class CustomError(Exception):
            pass

        def fail_callable(**kwargs):
            raise CustomError("Simulated task failure")

        _, task_instance = self._create_test_dag_and_task(fail_callable, "failure")
        # try_number before execution
        assert task_instance.try_number == TRY_NUMBER_BEFORE_EXECUTION
        with suppress(CustomError):
            task_instance.run()

        # try_number at the moment of function being called
        assert captured_try_numbers["running"] == TRY_NUMBER_RUNNING
        assert captured_try_numbers["failed"] == TRY_NUMBER_FAILED

        # try_number after task has been executed
        assert task_instance.try_number == TRY_NUMBER_AFTER_EXECUTION

    @mock.patch("airflow.models.taskinstance.get_listener_manager")
    def test_listener_on_task_instance_success_is_called_after_try_number_increment(self, mock_listener):
        """Validates the listener's on-success method is called before try_number increment happens.

        This test ensures that when a task instance successfully completes, the
        `on_task_instance_success` method of Airflow's listener is called with an
        incremented `try_number` compared to the `try_number` before execution.
        The test simulates a successful task execution and captures the `try_number` at the method call.
        """
        captured_try_numbers = {}
        self._setup_mock_listener(mock_listener, captured_try_numbers)

        def success_callable(**kwargs):
            return None

        _, task_instance = self._create_test_dag_and_task(success_callable, "success")
        # try_number before execution
        assert task_instance.try_number == TRY_NUMBER_BEFORE_EXECUTION
        task_instance.run()

        # try_number at the moment of function being called
        assert captured_try_numbers["running"] == TRY_NUMBER_RUNNING
        assert captured_try_numbers["success"] == TRY_NUMBER_SUCCESS

        # try_number after task has been executed
        assert task_instance.try_number == TRY_NUMBER_AFTER_EXECUTION

    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    def test_listener_on_task_instance_running_do_not_call_adapter_when_disabled_operator(
        self, mock_get_user_provided_run_facets, mock_get_airflow_run_facet, mock_disabled
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_user_provided_run_facets.return_value = {"custom_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_disabled.return_value = True

        listener.on_task_instance_running(None, task_instance, None)
        mock_disabled.assert_called_once_with(task_instance.task)
        listener.adapter.build_dag_run_id.assert_not_called()
        listener.adapter.build_task_instance_run_id.assert_not_called()
        listener.extractor_manager.extract_metadata.assert_not_called()
        listener.adapter.start_task.assert_not_called()

    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    def test_listener_on_task_instance_failed_do_not_call_adapter_when_disabled_operator(
        self, mock_get_user_provided_run_facets, mock_disabled
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_user_provided_run_facets.return_value = {"custom_facet": 2}
        mock_disabled.return_value = True

        on_task_failed_kwargs = {"error": ValueError("test")}

        listener.on_task_instance_failed(
            previous_state=None, task_instance=task_instance, **on_task_failed_kwargs, session=None
        )
        mock_disabled.assert_called_once_with(task_instance.task)
        listener.adapter.build_dag_run_id.assert_not_called()
        listener.adapter.build_task_instance_run_id.assert_not_called()
        listener.extractor_manager.extract_metadata.assert_not_called()
        listener.adapter.fail_task.assert_not_called()

    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    def test_listener_on_task_instance_success_do_not_call_adapter_when_disabled_operator(
        self, mock_get_user_provided_run_facets, mock_disabled
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_user_provided_run_facets.return_value = {"custom_facet": 2}
        mock_disabled.return_value = True

        listener.on_task_instance_success(None, task_instance, None)
        mock_disabled.assert_called_once_with(task_instance.task)
        listener.adapter.build_dag_run_id.assert_not_called()
        listener.adapter.build_task_instance_run_id.assert_not_called()
        listener.extractor_manager.extract_metadata.assert_not_called()
        listener.adapter.complete_task.assert_not_called()

    @pytest.mark.parametrize(
        "max_workers,expected",
        [
            (None, 1),
            ("8", 8),
        ],
    )
    @mock.patch("airflow.providers.openlineage.plugins.listener.ProcessPoolExecutor", autospec=True)
    def test_listener_on_dag_run_state_changes_configure_process_pool_size(
        self, mock_executor, max_workers, expected
    ):
        """mock ProcessPoolExecutor and check if conf.dag_state_change_process_pool_size is applied to max_workers"""
        listener = OpenLineageListener()
        # mock ProcessPoolExecutor class
        with conf_vars({("openlineage", "dag_state_change_process_pool_size"): max_workers}):
            listener.on_dag_run_running(mock.MagicMock(), None)
        mock_executor.assert_called_once_with(max_workers=expected, initializer=mock.ANY)
        mock_executor.return_value.submit.assert_called_once()

    @pytest.mark.db_test
    @pytest.mark.parametrize(
        ("method", "dag_run_state"),
        [
            ("on_dag_run_running", DagRunState.RUNNING),
            ("on_dag_run_success", DagRunState.SUCCESS),
            ("on_dag_run_failed", DagRunState.FAILED),
        ],
    )
    @patch("airflow.providers.openlineage.plugins.adapter.OpenLineageAdapter.emit")
    def test_listener_on_dag_run_state_changes(self, mock_emit, method, dag_run_state, create_task_instance):
        mock_executor = MockExecutor()
        ti = create_task_instance(dag_id="dag", task_id="op")
        # Change the state explicitly to set end_date following the logic in the method
        ti.dag_run.set_state(dag_run_state)
        with mock.patch(
            "airflow.providers.openlineage.plugins.listener.ProcessPoolExecutor", return_value=mock_executor
        ):
            listener = OpenLineageListener()
            getattr(listener, method)(ti.dag_run, None)
            assert mock_executor.submitted is True
            assert mock_executor.succeeded is True
            mock_emit.assert_called_once()

    def test_listener_logs_failed_serialization(self):
        listener = OpenLineageListener()
        callback_future = Future()

        def set_result(*args, **kwargs):
            callback_future.set_result(True)

        listener.log = MagicMock()
        listener.log.warning = MagicMock(side_effect=set_result)
        listener.adapter = OpenLineageAdapter(
            client=OpenLineageClient(transport=ConsoleTransport(config=ConsoleConfig()))
        )
        event_time = timezone.utcnow()
        fut = listener.submit_callable(
            listener.adapter.dag_failed,
            dag_id="",
            run_id="",
            end_date=event_time,
            logical_date=callback_future,
            clear_number=0,
            dag_run_state=DagRunState.FAILED,
            task_ids=["task_id"],
            msg="",
        )
        assert fut.exception(10)
        callback_future.result(10)
        assert callback_future.done()
        listener.log.debug.assert_not_called()
        listener.log.warning.assert_called_once()


@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Airflow 3 tests")
class TestOpenLineageListenerAirflow3:
    @pytest.mark.skip("Rendering fields is not migrated yet in Airflow 3")
    @patch("airflow.models.BaseOperator.render_template")
    def test_listener_does_not_change_task_instance(self, render_mock, mock_supervisor_comms, spy_agency):
        from airflow.sdk.execution_time.task_runner import (
            RuntimeTaskInstance,
            TaskInstance as SdkTaskInstance,
            run,
        )

        render_mock.return_value = render_df()

        date = timezone.datetime(2022, 1, 1)
        dag = DAG(
            "test",
            schedule=None,
            start_date=date,
            user_defined_macros={"render_df": render_df},
            params={"df": {"col": [1, 2]}},
        )
        task = TemplateOperator(task_id="template_op", dag=dag, do_xcom_push=True, df=dag.param("df"))
        run_id = str(uuid.uuid1())

        dagrun_kwargs = {
            "dag_version": None,
            "logical_date": date,
            "triggered_by": types.DagRunTriggeredByType.TEST,
        }

        dag.create_dagrun(
            run_id=run_id,
            data_interval=(date, date),
            run_type=types.DagRunType.MANUAL,
            state=DagRunState.QUEUED,
            **dagrun_kwargs,
        )
        ti = SdkTaskInstance(
            id=uuid7(),
            task_id="template_op",
            dag_id=dag.dag_id,
            run_id=run_id,
            try_number=1,
            map_index=-1,
            dag_version_id=uuid7(),
        )

        runtime_ti = RuntimeTaskInstance.model_construct(**ti.model_dump(exclude_unset=True), task=task)

        spy_agency.spy_on(runtime_ti.xcom_push, call_original=False)
        run(runtime_ti, None)

        # check if task returns the same DataFrame
        pd.testing.assert_frame_equal(runtime_ti.xcom_push.last_call.args[1], render_df())

        # check if render_template method always get the same unrendered field
        assert not isinstance(runtime_ti.xcom_push.last_call.args[1], pd.DataFrame)

    def _setup_mock_listener(self, mock_listener: mock.Mock, captured_try_numbers: dict[str, int]) -> None:
        """Sets up the mock listener with side effects to capture try numbers for different task instance events.

        :param mock_listener: The mock object for the listener manager.
        :param captured_try_numbers: A dictionary to store captured try numbers keyed by event names.

        This function iterates through specified event names and sets a side effect on the corresponding
        method of the listener manager's hook. The side effect is a nested function that captures the try number
        of the task instance when the method is called.

        :Example:

            captured_try_numbers = {}
            mock_listener = Mock()
            _setup_mock_listener(mock_listener, captured_try_numbers)
            # After running a task, captured_try_numbers will have the try number captured at the moment of
            execution for specified methods. F.e. {"running": 1, "success": 2} for on_task_instance_running and
            on_task_instance_success methods.
        """

        def capture_try_number(method_name):
            def inner(*args, **kwargs):
                captured_try_numbers[method_name] = kwargs["task_instance"].try_number

            return inner

        for event in ["running", "success", "failed"]:
            getattr(
                mock_listener.return_value.hook, f"on_task_instance_{event}"
            ).side_effect = capture_try_number(event)

    def _create_test_dag_and_task(
        self, python_callable: Callable, scenario_name: str
    ) -> tuple[DagRun, TaskInstance]:
        """Creates a test DAG and a task for a custom test scenario.

        :param python_callable: The Python callable to be executed by the PythonOperator.
        :param scenario_name: The name of the test scenario, used to uniquely name the DAG and task.

        :return: TaskInstance: The created TaskInstance object.

        This function creates a DAG and a PythonOperator task with the provided python_callable. It generates a unique
        run ID and creates a DAG run. This setup is useful for testing different scenarios in Airflow tasks.

        :Example:

            def sample_callable(**kwargs):
                print("Hello World")

            task_instance = _create_test_dag_and_task(sample_callable, "sample_scenario")
            # Use task_instance to simulate running a task in a test.
        """
        date = timezone.datetime(2022, 1, 1)
        dag = DAG(
            f"test_{scenario_name}",
            schedule=None,
            start_date=date,
        )
        t = PythonOperator(task_id=f"test_task_{scenario_name}", dag=dag, python_callable=python_callable)
        run_id = str(uuid.uuid1())
        dagrun_kwargs = {
            "dag_version": None,
            "logical_date": date,
            "triggered_by": types.DagRunTriggeredByType.TEST,
        }

        dagrun = dag.create_dagrun(
            run_id=run_id,
            data_interval=(date, date),
            start_date=date,
            run_type=types.DagRunType.MANUAL,
            state=DagRunState.QUEUED,
            **dagrun_kwargs,
        )
        task_instance = TaskInstance(t, run_id=run_id)  # type: ignore
        task_instance.dag_run = dagrun
        return dagrun, task_instance

    def _create_listener_and_task_instance(
        self, runtime_ti: bool = True
    ) -> tuple[OpenLineageListener, RuntimeTaskInstance | TaskInstance]:
        """Creates and configures an OpenLineageListener instance and a mock task instance for testing.

        :arg runtime_ti: Whether we should return mock RuntimeTaskInstance or mock TaskInstance
        :return: A tuple containing the configured OpenLineageListener and task instance.

        This function instantiates an OpenLineageListener, sets up its required properties with mock objects, and
        creates a mock task instance with predefined attributes. This setup is commonly used for testing the
        interaction between an OpenLineageListener and a task instance in Airflow.

        :Example:

            listener, task_instance = _create_listener_and_task_instance()
            # Now you can use listener and task_instance in your tests to simulate their interaction.
        """

        if not runtime_ti:
            # TaskInstance is used when on API server (when listener gets called about manual state change)
            task_instance = TaskInstance(task=MagicMock(), dag_version_id=uuid7())
            task_instance.dag_run = DagRun()
            task_instance.dag_run.dag_id = "dag_id_from_dagrun_and_not_ti"
            task_instance.dag_run.run_id = "dag_run_run_id"
            task_instance.dag_run.clear_number = 0
            task_instance.dag_run.logical_date = timezone.datetime(2020, 1, 1, 1, 1, 1)
            task_instance.dag_run.run_after = timezone.datetime(2020, 1, 1, 1, 1, 1)
            task_instance.dag_run.state = DagRunState.RUNNING
            task_instance.task = None
            task_instance.dag = None
            task_instance.task_id = "task_id"
            task_instance.dag_id = "dag_id"
            task_instance.try_number = 1
            task_instance.map_index = -1
        else:
            # RuntimeTaskInstance is used when on worker
            from airflow.sdk.api.datamodels._generated import (
                DagRun as SdkDagRun,
                DagRunState as SdkDagRunState,
                DagRunType,
                TaskInstance as SdkTaskInstance,
                TIRunContext,
            )
            from airflow.sdk.definitions.dag import DAG
            from airflow.sdk.execution_time.task_runner import RuntimeTaskInstance

            dag = DAG(
                dag_id="dag_id_from_dag_not_ti",
                description="Test DAG Description",
                tags=["tag1", "tag2"],
            )
            task = EmptyOperator(
                task_id="task_id_from_task_not_ti", dag=dag, owner="task_owner", doc_md="TASK Description"
            )
            task2 = EmptyOperator(task_id="task_id2_from_task_not_ti", dag=dag, owner="another_owner")  # noqa: F841

            sdk_task_instance = SdkTaskInstance(
                id=uuid7(),
                task_id="task_id",
                dag_id="dag_id",
                run_id="dag_run_run_id",
                try_number=1,
                map_index=-1,
                dag_version_id=uuid7(),
            )
            task_instance = RuntimeTaskInstance.model_construct(  # type: ignore
                **sdk_task_instance.model_dump(exclude_unset=True),
                task=task,
                _ti_context_from_server=TIRunContext(
                    dag_run=SdkDagRun.model_validate(
                        {
                            "dag_id": "dag_id_from_dagrun_not_ti",
                            "run_id": "dag_run_run_id_from_dagrun_not_ti",
                            "logical_date": timezone.datetime(2020, 1, 1, 1, 1, 1),
                            "start_date": timezone.datetime(2023, 1, 1, 13, 1, 1),
                            "end_date": timezone.datetime(2023, 1, 3, 13, 1, 1),
                            "run_type": DagRunType.MANUAL,
                            "run_after": timezone.datetime(2023, 1, 3, 13, 1, 1),
                            "consumed_asset_events": [],
                            **(
                                {"state": SdkDagRunState.RUNNING} if "state" in SdkDagRun.model_fields else {}
                            ),
                        }
                    ),
                    task_reschedule_count=0,
                    max_tries=1,
                    should_retry=False,
                ),
                start_date=timezone.datetime(2023, 1, 1, 13, 1, 1),
            )

        def mock_dag_id(dag_id, logical_date, clear_number):
            return f"{logical_date.isoformat()}.{dag_id}.{clear_number}"

        def mock_task_id(dag_id, task_id, try_number, logical_date, map_index):
            return f"{logical_date.isoformat()}.{dag_id}.{task_id}.{try_number}.{map_index}"

        listener = OpenLineageListener()
        listener.extractor_manager = mock.Mock()

        metadata = mock.Mock()
        metadata.run_facets = {"run_facet": 1}
        listener.extractor_manager.extract_metadata.return_value = metadata

        adapter = mock.Mock()
        adapter.build_dag_run_id.side_effect = mock_dag_id
        adapter.build_task_instance_run_id.side_effect = mock_task_id
        adapter.start_task = mock.Mock()
        adapter.fail_task = mock.Mock()
        adapter.complete_task = mock.Mock()
        listener.adapter = adapter

        return listener, task_instance

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_mapped_task_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_start_task_is_called_with_proper_arguments(
        self,
        mock_get_airflow_mapped_task_facet,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
    ):
        """Tests that the 'start_task' method of the OpenLineageAdapter is invoked with the correct arguments.

        The test checks that the job name, job description, event time, and other related data are
        correctly passed to the adapter. It also verifies that custom facets and Airflow run facets are
        correctly retrieved and included in the call. This ensures that all relevant data, including custom
        and Airflow-specific metadata, is accurately conveyed to the adapter during the initialization of a task,
        reflecting the comprehensive tracking of task execution contexts."""

        listener, task_instance = self._create_listener_and_task_instance()

        mock_get_airflow_mapped_task_facet.return_value = {"mapped_facet": 1}
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        listener.on_task_instance_running(None, task_instance)
        listener.adapter.start_task.assert_called_once_with(
            run_id="2020-01-01T01:01:01+00:00.dag_id.task_id.1.-1",
            job_name="dag_id.task_id",
            job_description="TASK Description",
            job_description_type="text/markdown",
            event_time="2023-01-01T13:01:01+00:00",
            nominal_start_time=None,
            nominal_end_time=None,
            owners=["task_owner"],
            tags={"tag1", "tag2"},
            task=listener.extractor_manager.extract_metadata(),
            run_facets={
                "mapped_facet": 1,
                "custom_user_facet": 2,
                "airflow_run_facet": 3,
                "parent": 4,
                "debug": "packages",
            },
        )

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_mapped_task_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_start_task_is_called_with_dag_owners_when_task_owner_is_default(
        self,
        mock_get_airflow_mapped_task_facet,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_airflow_mapped_task_facet.return_value = {"mapped_facet": 1}
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        task_instance.task.owner = "airflow"  # Simulate default owner on task to force fallback to DAG owner
        listener.on_task_instance_running(None, task_instance)
        call_owners = listener.adapter.start_task.call_args.kwargs["owners"]
        assert sorted(call_owners) == ["airflow", "another_owner"]

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_mapped_task_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_job_name")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_start_task_is_called_with_dag_description_when_task_doc_is_empty(
        self,
        mock_get_job_name,
        mock_get_airflow_mapped_task_facet,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_job_name.return_value = "job_name"
        mock_get_airflow_mapped_task_facet.return_value = {"mapped_facet": 1}
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        task_instance.task.doc_md = None  # Simulate lack of task doc to force fallback to DAG description
        listener.on_task_instance_running(None, task_instance)
        assert listener.adapter.start_task.call_args.kwargs["job_description"] == "Test DAG Description"
        assert listener.adapter.start_task.call_args.kwargs["job_description_type"] == "text/plain"

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_fail_task_is_called_with_proper_arguments(
        self,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
        time_machine,
    ):
        """Tests that the 'fail_task' method of the OpenLineageAdapter is invoked with the correct arguments.

        This test ensures that the job name is accurately retrieved and included, along with the generated
        run_id and task metadata. By mocking the job name retrieval and the run_id generation,
        the test verifies the integrity and consistency of the data passed to the adapter during task
        failure events, thus confirming that the adapter's failure handling is functioning as expected.
        """
        time_machine.move_to(timezone.datetime(2023, 1, 3, 13, 1, 1), tick=False)

        listener, task_instance = self._create_listener_and_task_instance()
        task_instance.get_template_context()["dag_run"].logical_date = timezone.datetime(2020, 1, 1, 1, 1, 1)
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow": {"task": "..."}}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        err = ValueError("test")
        listener.on_task_instance_failed(previous_state=None, task_instance=task_instance, error=err)
        listener.adapter.fail_task.assert_called_once_with(
            end_time="2023-01-03T13:01:01+00:00",
            job_name="dag_id.task_id",
            run_id="2020-01-01T01:01:01+00:00.dag_id.task_id.1.-1",
            task=listener.extractor_manager.extract_metadata(),
            owners=["task_owner"],
            tags={"tag1", "tag2"},
            job_description="TASK Description",
            job_description_type="text/markdown",
            nominal_start_time=None,
            nominal_end_time=None,
            run_facets={
                "parent": 4,
                "custom_user_facet": 2,
                "airflow": {"task": "..."},
                "debug": "packages",
            },
            error=err,
        )

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_mapped_task_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_fail_task_is_called_with_dag_owners_when_task_owner_is_default(
        self,
        mock_get_airflow_mapped_task_facet,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_airflow_mapped_task_facet.return_value = {"mapped_facet": 1}
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        task_instance.task.owner = "airflow"  # Simulate default owner on task to force fallback to DAG owner
        err = ValueError("test")
        listener.on_task_instance_failed(previous_state=None, task_instance=task_instance, error=err)

        call_owners = listener.adapter.fail_task.call_args.kwargs["owners"]
        assert sorted(call_owners) == ["airflow", "another_owner"]

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_mapped_task_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_fail_task_is_called_with_dag_description_when_task_doc_is_empty(
        self,
        mock_get_airflow_mapped_task_facet,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_airflow_mapped_task_facet.return_value = {"mapped_facet": 1}
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        task_instance.task.doc_md = None  # Simulate lack of task doc to force fallback to DAG description
        err = ValueError("test")
        listener.on_task_instance_failed(previous_state=None, task_instance=task_instance, error=err)
        assert listener.adapter.fail_task.call_args.kwargs["job_description"] == "Test DAG Description"
        assert listener.adapter.fail_task.call_args.kwargs["job_description_type"] == "text/plain"

    @mock.patch("airflow.providers.openlineage.plugins.adapter.OpenLineageAdapter.emit")
    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_fail_task_is_called_with_proper_arguments_for_db_task_instance_model(
        self,
        mock_get_task_parent_run_facet,
        mock_debug_facet,
        mock_debug_mode,
        mock_emit,
        time_machine,
    ):
        """Tests that the 'fail_task' method of the OpenLineageAdapter is invoked with the correct arguments.

        This particular test is using TaskInstance model available on API Server and not on worker,
        to simulate the listener being called after task's state has been manually set via API.
        """
        time_machine.move_to(timezone.datetime(2023, 1, 3, 13, 1, 1), tick=False)

        listener, task_instance = self._create_listener_and_task_instance(runtime_ti=False)
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}

        err = ValueError("test")

        listener.on_task_instance_failed(previous_state=None, task_instance=task_instance, error=err)
        mock_get_task_parent_run_facet.assert_called_once_with(
            parent_run_id="2020-01-01T01:01:01+00:00.dag_id.0", parent_job_name=task_instance.dag_id
        )
        expected_args = dict(
            end_time="2023-01-03T13:01:01+00:00",
            job_name="dag_id.task_id",
            run_id="2020-01-01T01:01:01+00:00.dag_id.task_id.1.-1",
            task=OperatorLineage(),
            nominal_start_time=None,
            nominal_end_time=None,
            tags=None,
            owners=None,
            job_description=None,
            job_description_type=None,
            run_facets={
                "parent": 4,
                "debug": "packages",
            },
            error=err,
        )
        listener.adapter.fail_task.assert_called_once_with(**expected_args)

        expected_args["run_id"] = "9d3b14f7-de91-40b6-aeef-e887e2c7673e"
        adapter = OpenLineageAdapter()
        adapter.fail_task(**expected_args)
        assert mock_emit.assert_called_once

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_complete_task_is_called_with_proper_arguments(
        self,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
        time_machine,
    ):
        """Tests that the 'complete_task' method of the OpenLineageAdapter is called with the correct arguments.

        It checks that the job name is correctly retrieved and passed,
        along with the run_id and task metadata. The test also simulates changes in the try_number
        attribute of the task instance, as it would occur in Airflow, to ensure that the run_id is updated
        accordingly. This helps confirm the consistency and correctness of the data passed to the adapter
        during the task's lifecycle events.
        """
        time_machine.move_to(timezone.datetime(2023, 1, 3, 13, 1, 1), tick=False)

        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow": {"task": "..."}}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        listener.on_task_instance_success(None, task_instance)
        calls = listener.adapter.complete_task.call_args_list
        assert len(calls) == 1
        assert calls[0][1] == dict(
            end_time="2023-01-03T13:01:01+00:00",
            job_name="dag_id.task_id",
            run_id="2020-01-01T01:01:01+00:00.dag_id.task_id.1.-1",
            task=listener.extractor_manager.extract_metadata(),
            owners=["task_owner"],
            tags={"tag1", "tag2"},
            job_description="TASK Description",
            job_description_type="text/markdown",
            nominal_start_time=None,
            nominal_end_time=None,
            run_facets={
                "parent": 4,
                "custom_user_facet": 2,
                "airflow": {"task": "..."},
                "debug": "packages",
            },
        )

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_mapped_task_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_complete_task_is_called_with_dag_owners_when_task_owner_is_default(
        self,
        mock_get_airflow_mapped_task_facet,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_airflow_mapped_task_facet.return_value = {"mapped_facet": 1}
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        task_instance.task.owner = "airflow"  # Simulate default owner on task to force fallback to DAG owner
        listener.on_task_instance_success(None, task_instance)

        call_owners = listener.adapter.complete_task.call_args.kwargs["owners"]
        assert sorted(call_owners) == ["airflow", "another_owner"]

    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_mapped_task_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_complete_task_is_called_with_dag_description_when_task_doc_is_empty(
        self,
        mock_get_airflow_mapped_task_facet,
        mock_get_user_provided_run_facets,
        mock_get_airflow_run_facet,
        mock_get_task_parent_run_facet,
        mock_disabled,
        mock_debug_facet,
        mock_debug_mode,
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_airflow_mapped_task_facet.return_value = {"mapped_facet": 1}
        mock_get_user_provided_run_facets.return_value = {"custom_user_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}
        mock_disabled.return_value = False

        task_instance.task.doc_md = None  # Simulate lack of task doc to force fallback to DAG description
        listener.on_task_instance_success(None, task_instance)
        assert listener.adapter.complete_task.call_args.kwargs["job_description"] == "Test DAG Description"
        assert listener.adapter.complete_task.call_args.kwargs["job_description_type"] == "text/plain"

    @mock.patch("airflow.providers.openlineage.plugins.adapter.OpenLineageAdapter.emit")
    @mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_debug_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_task_parent_run_facet")
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_adapter_complete_task_is_called_with_proper_arguments_for_db_task_instance_model(
        self, mock_get_task_parent_run_facet, mock_debug_facet, mock_debug_mode, mock_emit, time_machine
    ):
        """Tests that the 'complete_task' method of the OpenLineageAdapter is called with the correct arguments.

        This particular test is using TaskInstance model available on API Server and not on worker,
        to simulate the listener being called after task's state has been manually set via API.
        """
        time_machine.move_to(timezone.datetime(2023, 1, 3, 13, 1, 1), tick=False)

        listener, task_instance = self._create_listener_and_task_instance(runtime_ti=False)
        mock_get_task_parent_run_facet.return_value = {"parent": 4}
        mock_debug_facet.return_value = {"debug": "packages"}

        listener.on_task_instance_success(None, task_instance)
        calls = listener.adapter.complete_task.call_args_list
        assert len(calls) == 1
        mock_get_task_parent_run_facet.assert_called_once_with(
            parent_run_id="2020-01-01T01:01:01+00:00.dag_id.0", parent_job_name=task_instance.dag_id
        )
        expected_args = dict(
            end_time="2023-01-03T13:01:01+00:00",
            job_name="dag_id.task_id",
            run_id="2020-01-01T01:01:01+00:00.dag_id.task_id.1.-1",
            task=OperatorLineage(),
            nominal_start_time=None,
            nominal_end_time=None,
            tags=None,
            owners=None,
            job_description=None,
            job_description_type=None,
            run_facets={
                "parent": 4,
                "debug": "packages",
            },
        )
        assert calls[0][1] == expected_args

        expected_args["run_id"] = "9d3b14f7-de91-40b6-aeef-e887e2c7673e"
        adapter = OpenLineageAdapter()
        adapter.complete_task(**expected_args)
        assert mock_emit.assert_called_once

    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_on_task_instance_running_correctly_calls_openlineage_adapter_run_id_method(self):
        """Tests the OpenLineageListener's response when a task instance is in the running state.

        This test ensures that when an Airflow task instance transitions to the running state,
        the OpenLineageAdapter's `build_task_instance_run_id` method is called exactly once with the correct
        parameters derived from the task instance.
        """
        listener, task_instance = self._create_listener_and_task_instance()
        listener.on_task_instance_running(None, task_instance)
        listener.adapter.build_task_instance_run_id.assert_called_once_with(
            dag_id="dag_id",
            task_id="task_id",
            logical_date=timezone.datetime(2020, 1, 1, 1, 1, 1),
            try_number=1,
            map_index=-1,
        )

    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_on_task_instance_failed_correctly_calls_openlineage_adapter_run_id_method(self):
        """Tests the OpenLineageListener's response when a task instance is in the failed state.

        This test ensures that when an Airflow task instance transitions to the failed state,
        the OpenLineageAdapter's `build_task_instance_run_id` method is called exactly once with the correct
        parameters derived from the task instance.
        """
        listener, task_instance = self._create_listener_and_task_instance()
        on_task_failed_kwargs = {"error": ValueError("test")}

        listener.on_task_instance_failed(
            previous_state=None, task_instance=task_instance, **on_task_failed_kwargs
        )
        listener.adapter.build_task_instance_run_id.assert_called_once_with(
            dag_id="dag_id",
            task_id="task_id",
            logical_date=timezone.datetime(2020, 1, 1, 1, 1, 1),
            try_number=1,
            map_index=-1,
        )

    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_on_task_instance_success_correctly_calls_openlineage_adapter_run_id_method(self):
        """Tests the OpenLineageListener's response when a task instance is in the success state.

        This test ensures that when an Airflow task instance transitions to the success state,
        the OpenLineageAdapter's `build_task_instance_run_id` method is called exactly once with the correct
        parameters derived from the task instance.
        """
        listener, task_instance = self._create_listener_and_task_instance()
        listener.on_task_instance_success(None, task_instance)
        listener.adapter.build_task_instance_run_id.assert_called_once_with(
            dag_id="dag_id",
            task_id="task_id",
            logical_date=timezone.datetime(2020, 1, 1, 1, 1, 1),
            try_number=1,
            map_index=-1,
        )

    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_airflow_run_facet")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    def test_listener_on_task_instance_running_do_not_call_adapter_when_disabled_operator(
        self, mock_get_user_provided_run_facets, mock_get_airflow_run_facet, mock_disabled
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_user_provided_run_facets.return_value = {"custom_facet": 2}
        mock_get_airflow_run_facet.return_value = {"airflow_run_facet": 3}
        mock_disabled.return_value = True

        listener.on_task_instance_running(None, task_instance)
        mock_disabled.assert_called_once_with(task_instance.task)
        listener.adapter.build_dag_run_id.assert_not_called()
        listener.adapter.build_task_instance_run_id.assert_not_called()
        listener.extractor_manager.extract_metadata.assert_not_called()
        listener.adapter.start_task.assert_not_called()

    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    def test_listener_on_task_instance_failed_do_not_call_adapter_when_disabled_operator(
        self, mock_get_user_provided_run_facets, mock_disabled
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_user_provided_run_facets.return_value = {"custom_facet": 2}
        mock_disabled.return_value = True

        on_task_failed_kwargs = {"error": ValueError("test")}

        listener.on_task_instance_failed(
            previous_state=None, task_instance=task_instance, **on_task_failed_kwargs
        )
        mock_disabled.assert_called_once_with(task_instance.task)
        listener.adapter.build_dag_run_id.assert_not_called()
        listener.adapter.build_task_instance_run_id.assert_not_called()
        listener.extractor_manager.extract_metadata.assert_not_called()
        listener.adapter.fail_task.assert_not_called()

    @mock.patch("airflow.providers.openlineage.plugins.listener.is_operator_disabled")
    @mock.patch("airflow.providers.openlineage.plugins.listener.get_user_provided_run_facets")
    def test_listener_on_task_instance_success_do_not_call_adapter_when_disabled_operator(
        self, mock_get_user_provided_run_facets, mock_disabled
    ):
        listener, task_instance = self._create_listener_and_task_instance()
        mock_get_user_provided_run_facets.return_value = {"custom_facet": 2}
        mock_disabled.return_value = True

        listener.on_task_instance_success(None, task_instance)
        mock_disabled.assert_called_once_with(task_instance.task)
        listener.adapter.build_dag_run_id.assert_not_called()
        listener.adapter.build_task_instance_run_id.assert_not_called()
        listener.extractor_manager.extract_metadata.assert_not_called()
        listener.adapter.complete_task.assert_not_called()

    @pytest.mark.parametrize(
        "max_workers,expected",
        [
            (None, 1),
            ("8", 8),
        ],
    )
    @mock.patch("airflow.providers.openlineage.plugins.listener.ProcessPoolExecutor", autospec=True)
    def test_listener_on_dag_run_state_changes_configure_process_pool_size(
        self, mock_executor, max_workers, expected
    ):
        """mock ProcessPoolExecutor and check if conf.dag_state_change_process_pool_size is applied to max_workers"""
        listener = OpenLineageListener()
        # mock ProcessPoolExecutor class
        with conf_vars({("openlineage", "dag_state_change_process_pool_size"): max_workers}):
            listener.on_dag_run_running(mock.MagicMock(), None)
        mock_executor.assert_called_once_with(max_workers=expected, initializer=mock.ANY)
        mock_executor.return_value.submit.assert_called_once()

    @pytest.mark.db_test
    @pytest.mark.parametrize(
        ("method", "dag_run_state"),
        [
            ("on_dag_run_running", DagRunState.RUNNING),
            ("on_dag_run_success", DagRunState.SUCCESS),
            ("on_dag_run_failed", DagRunState.FAILED),
        ],
    )
    @patch("airflow.providers.openlineage.plugins.adapter.OpenLineageAdapter.emit")
    def test_listener_on_dag_run_state_changes(self, mock_emit, method, dag_run_state, create_task_instance):
        mock_executor = MockExecutor()
        ti = create_task_instance(dag_id="dag", task_id="op")
        # Change the state explicitly to set end_date following the logic in the method
        ti.dag_run.set_state(dag_run_state)
        with mock.patch(
            "airflow.providers.openlineage.plugins.listener.ProcessPoolExecutor", return_value=mock_executor
        ):
            listener = OpenLineageListener()
            getattr(listener, method)(ti.dag_run, None)
            assert mock_executor.submitted is True
            assert mock_executor.succeeded is True
            mock_emit.assert_called_once()

    def test_listener_logs_failed_serialization(self):
        listener = OpenLineageListener()
        callback_future = Future()

        def set_result(*args, **kwargs):
            callback_future.set_result(True)

        listener.log = MagicMock()
        listener.log.warning = MagicMock(side_effect=set_result)
        listener.adapter = OpenLineageAdapter(
            client=OpenLineageClient(transport=ConsoleTransport(config=ConsoleConfig()))
        )
        event_time = timezone.utcnow()
        fut = listener.submit_callable(
            listener.adapter.dag_failed,
            dag_id="",
            run_id="",
            end_date=event_time,
            logical_date=callback_future,
            clear_number=0,
            dag_run_state=DagRunState.FAILED,
            task_ids=["task_id"],
            msg="",
        )
        assert fut.exception(10)
        callback_future.result(10)
        assert callback_future.done()
        listener.log.debug.assert_not_called()
        listener.log.warning.assert_called_once()


@pytest.mark.skipif(AIRFLOW_V_3_0_PLUS, reason="Airflow 2 tests")
class TestOpenLineageSelectiveEnableAirflow2:
    def setup_method(self):
        date = timezone.datetime(2022, 1, 1)
        self.dag = DAG(
            "test_selective_enable",
            schedule=None,
            start_date=date,
        )

        def simple_callable(**kwargs):
            return None

        self.task_1 = PythonOperator(
            task_id="test_task_selective_enable_1", dag=self.dag, python_callable=simple_callable
        )
        self.task_2 = PythonOperator(
            task_id="test_task_selective_enable_2", dag=self.dag, python_callable=simple_callable
        )
        run_id = str(uuid.uuid1())
        self.dagrun = self.dag.create_dagrun(
            run_id=run_id,
            data_interval=(date, date),
            run_type=types.DagRunType.MANUAL,
            state=DagRunState.QUEUED,
            execution_date=date,
        )
        self.task_instance_1 = TaskInstance(self.task_1, run_id=run_id, map_index=-1)
        self.task_instance_2 = TaskInstance(self.task_2, run_id=run_id, map_index=-1)
        self.task_instance_1.dag_run = self.task_instance_2.dag_run = self.dagrun

    def teardown_method(self):
        clear_db_runs()

    @pytest.mark.parametrize(
        "selective_enable, enable_dag, expected_call_count",
        [
            ("True", True, 3),
            ("False", True, 3),
            ("True", False, 0),
            ("False", False, 3),
        ],
    )
    def test_listener_with_dag_enabled(self, selective_enable, enable_dag, expected_call_count):
        """Tests listener's behaviour with selective-enable on DAG level."""

        if enable_dag:
            enable_lineage(self.dag)

        with conf_vars({("openlineage", "selective_enable"): selective_enable}):
            listener = OpenLineageListener()
            listener._executor = mock.Mock()

            # run all three DagRun-related hooks
            listener.on_dag_run_running(self.dagrun, msg="test running")
            listener.on_dag_run_failed(self.dagrun, msg="test failure")
            listener.on_dag_run_success(self.dagrun, msg="test success")

    @pytest.mark.parametrize(
        "selective_enable, enable_task, expected_dag_call_count, expected_task_call_count",
        [
            ("True", True, 3, 3),
            ("False", True, 3, 3),
            ("True", False, 0, 0),
            ("False", False, 3, 3),
        ],
    )
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_listener_with_task_enabled(
        self, selective_enable, enable_task, expected_dag_call_count, expected_task_call_count
    ):
        """Tests listener's behaviour with selective-enable on task level."""

        if enable_task:
            enable_lineage(self.task_1)

        on_task_failed_kwargs = {"error": ValueError("test")}

        with conf_vars({("openlineage", "selective_enable"): selective_enable}):
            listener = OpenLineageListener()
            listener._executor = mock.Mock()
            listener.extractor_manager = mock.Mock()
            listener.adapter = mock.Mock()

            # run all three DagRun-related hooks
            listener.on_dag_run_running(self.dagrun, msg="test running")
            listener.on_dag_run_failed(self.dagrun, msg="test failure")
            listener.on_dag_run_success(self.dagrun, msg="test success")

            assert expected_dag_call_count == listener._executor.submit.call_count

            # run TaskInstance-related hooks for lineage enabled task
            listener.on_task_instance_running(None, self.task_instance_1, None)
            listener.on_task_instance_success(None, self.task_instance_1, None)
            listener.on_task_instance_failed(
                previous_state=None, task_instance=self.task_instance_1, **on_task_failed_kwargs, session=None
            )

            assert expected_task_call_count == listener.extractor_manager.extract_metadata.call_count

            # run TaskInstance-related hooks for lineage disabled task
            listener.on_task_instance_running(None, self.task_instance_2, None)
            listener.on_task_instance_success(None, self.task_instance_2, None)
            listener.on_task_instance_failed(
                previous_state=None, task_instance=self.task_instance_2, **on_task_failed_kwargs, session=None
            )

            # with selective-enable disabled both task_1 and task_2 should trigger metadata extraction
            if selective_enable == "False":
                expected_task_call_count *= 2

            assert expected_task_call_count == listener.extractor_manager.extract_metadata.call_count

    @pytest.mark.parametrize(
        "selective_enable, enable_task, expected_call_count, expected_task_call_count",
        [
            ("True", True, 3, 3),
            ("False", True, 3, 3),
            ("True", False, 0, 0),
            ("False", False, 3, 3),
        ],
    )
    @mock.patch(
        "airflow.providers.openlineage.plugins.listener.OpenLineageListener._execute", new=regular_call
    )
    def test_listener_with_dag_disabled_task_enabled(
        self, selective_enable, enable_task, expected_call_count, expected_task_call_count
    ):
        """Tests listener's behaviour with selective-enable on task level with DAG disabled."""
        disable_lineage(self.dag)

        if enable_task:
            enable_lineage(self.task_1)

        on_task_failed_kwargs = {"error": ValueError("test")}

        with conf_vars({("openlineage", "selective_enable"): selective_enable}):
            listener = OpenLineageListener()
            listener._executor = mock.Mock()
            listener.extractor_manager = mock.Mock()
            listener.adapter = mock.Mock()

            # run all three DagRun-related hooks
            listener.on_dag_run_running(self.dagrun, msg="test running")
            listener.on_dag_run_failed(self.dagrun, msg="test failure")
            listener.on_dag_run_success(self.dagrun, msg="test success")

            # run TaskInstance-related hooks for lineage enabled task
            listener.on_task_instance_running(None, self.task_instance_1, session=None)
            listener.on_task_instance_success(None, self.task_instance_1, session=None)
            listener.on_task_instance_failed(
                previous_state=None, task_instance=self.task_instance_1, **on_task_failed_kwargs, session=None
            )

        assert expected_call_count == listener._executor.submit.call_count
        assert expected_task_call_count == listener.extractor_manager.extract_metadata.call_count



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
import json
import uuid
from json import JSONEncoder
from typing import TYPE_CHECKING, Any
from unittest.mock import MagicMock, patch

import pytest
from attrs import define
from openlineage.client.utils import RedactMixin
from pkg_resources import parse_version

from airflow.models import DAG, DagModel
from airflow.providers.common.compat.assets import Asset
from airflow.providers.openlineage.plugins.facets import AirflowDebugRunFacet
from airflow.providers.openlineage.utils.utils import (
    DagInfo,
    InfoJsonEncodable,
    OpenLineageRedactor,
    _get_all_packages_installed,
    _is_name_redactable,
    get_airflow_debug_facet,
    get_airflow_run_facet,
    get_fully_qualified_class_name,
    get_processing_engine_facet,
    is_operator_disabled,
)
from airflow.serialization.enums import DagAttributeTypes, Encoding
from airflow.utils import timezone
from airflow.utils.state import State
from airflow.utils.types import DagRunType

from tests_common.test_utils.compat import (
    BashOperator,
)
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.utils.types import DagRunTriggeredByType

if TYPE_CHECKING:
    from airflow.sdk.execution_time.secrets_masker import _secrets_masker
else:
    try:
        from airflow.sdk.execution_time.secrets_masker import _secrets_masker
    except ImportError:
        from airflow.utils.log.secrets_masker import _secrets_masker


class SafeStrDict(dict):
    def __str__(self):
        castable = []
        for key, val in self.items():
            try:
                str(key), str(val)
                castable.append((key, val))
            except (TypeError, NotImplementedError):
                continue
        return str(dict(castable))


@patch("airflow.providers.openlineage.utils.utils.metadata.distributions")
def test_get_all_packages_installed(mock_distributions):
    mock_distributions.return_value = [MagicMock(metadata={"Name": "package1"}, version="1.0.0")]
    assert _get_all_packages_installed() == {"package1": "1.0.0"}


@patch("airflow.providers.openlineage.utils.utils.conf.debug_mode", return_value=False)
def test_get_airflow_debug_facet_not_in_debug_mode(mock_debug_mode):
    assert get_airflow_debug_facet() == {}


@patch("airflow.providers.openlineage.utils.utils._get_all_packages_installed")
@patch("airflow.providers.openlineage.utils.utils.conf.debug_mode")
def test_get_airflow_debug_facet_logging_set_to_debug(mock_debug_mode, mock_get_packages):
    mock_debug_mode.return_value = True
    mock_get_packages.return_value = {"package1": "1.0.0"}
    result = get_airflow_debug_facet()
    expected_result = {"debug": AirflowDebugRunFacet(packages={"package1": "1.0.0"})}
    assert result == expected_result


@pytest.mark.db_test
def test_get_dagrun_start_end(dag_maker):
    start_date = datetime.datetime(2022, 1, 1)
    end_date = datetime.datetime(2022, 1, 1, hour=2)
    with dag_maker("test", start_date=start_date, end_date=end_date, schedule="@once") as dag:
        pass
    dag_maker.sync_dagbag_to_db()
    dag_model = DagModel.get_dagmodel(dag.dag_id)

    run_id = str(uuid.uuid1())
    data_interval = dag.get_next_data_interval(dag_model)
    if AIRFLOW_V_3_0_PLUS:
        dagrun_kwargs = {
            "logical_date": data_interval.start,
            "run_after": data_interval.end,
            "triggered_by": DagRunTriggeredByType.TEST,
        }
    else:
        dagrun_kwargs = {"execution_date": data_interval.start}
    dagrun = dag.create_dagrun(
        state=State.NONE,
        run_id=run_id,
        run_type=DagRunType.MANUAL,
        data_interval=data_interval,
        **dagrun_kwargs,
    )
    assert dagrun.data_interval_start is not None
    start_date_tz = datetime.datetime(2022, 1, 1, tzinfo=timezone.utc)
    end_date_tz = datetime.datetime(2022, 1, 1, hour=2, tzinfo=timezone.utc)
    assert dagrun.data_interval_start, dagrun.data_interval_end == (start_date_tz, end_date_tz)


def test_parse_version():
    assert parse_version("2.3.0") >= parse_version("2.3.0.dev0")
    assert parse_version("2.3.0.dev0") >= parse_version("2.3.0.dev0")
    assert parse_version("2.3.0.beta1") >= parse_version("2.3.0.dev0")
    assert parse_version("2.3.1") >= parse_version("2.3.0.dev0")
    assert parse_version("2.4.0") >= parse_version("2.3.0.dev0")
    assert parse_version("3.0.0") >= parse_version("2.3.0.dev0")
    assert parse_version("2.2.0") < parse_version("2.3.0.dev0")
    assert parse_version("2.1.3") < parse_version("2.3.0.dev0")
    assert parse_version("2.2.4") < parse_version("2.3.0.dev0")
    assert parse_version("1.10.15") < parse_version("2.3.0.dev0")
    assert parse_version("2.2.4.dev0") < parse_version("2.3.0.dev0")


def test_safe_dict():
    assert str(SafeStrDict({"a": 1})) == str({"a": 1})

    class NotImplemented:
        def __str__(self):
            raise NotImplementedError

    assert str(SafeStrDict({"a": NotImplemented()})) == str({})


def test_info_json_encodable():
    class TestInfo(InfoJsonEncodable):
        excludes = ["exclude_1", "exclude_2", "imastring"]
        casts = {"iwanttobeint": lambda x: int(x.imastring)}
        renames = {"_faulty_name": "goody_name"}

    @define
    class Test:
        exclude_1: str
        imastring: str
        _faulty_name: str
        donotcare: str

    obj = Test("val", "123", "not_funny", "abc")

    assert json.loads(json.dumps(TestInfo(obj))) == {
        "iwanttobeint": 123,
        "goody_name": "not_funny",
        "donotcare": "abc",
    }


def test_info_json_encodable_without_slots():
    class TestInfo(InfoJsonEncodable):
        excludes = ["exclude_1", "exclude_2", "imastring"]
        casts = {"iwanttobeint": lambda x: int(x.imastring)}
        renames = {"_faulty_name": "goody_name"}

    @define(slots=False)
    class Test:
        exclude_1: str
        imastring: str
        _faulty_name: str
        donotcare: str

    obj = Test("val", "123", "not_funny", "abc")

    assert json.loads(json.dumps(TestInfo(obj))) == {
        "iwanttobeint": 123,
        "goody_name": "not_funny",
        "donotcare": "abc",
    }


def test_info_json_encodable_list_does_flatten():
    class TestInfo(InfoJsonEncodable):
        includes = ["alist"]

    @define
    class Test:
        alist: list[str]

    obj = Test(["a", "b", "c"])

    assert json.loads(json.dumps(TestInfo(obj))) == {"alist": "['a', 'b', 'c']"}


def test_info_json_encodable_list_does_include_nonexisting():
    class TestInfo(InfoJsonEncodable):
        includes = ["exists", "doesnotexist"]

    @define
    class Test:
        exists: str

    obj = Test("something")

    assert json.loads(json.dumps(TestInfo(obj))) == {"exists": "something"}


def test_is_name_redactable():
    class NotMixin:
        def __init__(self):
            self.password = "passwd"

    class Mixined(RedactMixin):
        _skip_redact = ["password"]

        def __init__(self):
            self.password = "passwd"
            self.transparent = "123"

    assert _is_name_redactable("password", NotMixin())
    assert not _is_name_redactable("password", Mixined())
    assert _is_name_redactable("transparent", Mixined())


@pytest.mark.enable_redact
def test_redact_with_exclusions(monkeypatch):
    redactor = OpenLineageRedactor.from_masker(_secrets_masker())

    class NotMixin:
        def __init__(self):
            self.password = "passwd"

    class Proxy:
        pass

    def default(self, o):
        if isinstance(o, NotMixin):
            return o.__dict__
        raise TypeError

    assert redactor.redact(NotMixin()).password == "passwd"
    monkeypatch.setattr(JSONEncoder, "default", default)
    assert redactor.redact(NotMixin()).password == "***"

    assert redactor.redact(Proxy()) == "<<non-redactable: Proxy>>"
    assert redactor.redact({"a": "a", "b": Proxy()}) == {"a": "a", "b": "<<non-redactable: Proxy>>"}

    class Mixined(RedactMixin):
        _skip_redact = ["password"]

        def __init__(self):
            self.password = "passwd"
            self.transparent = "123"

    @define
    class NestedMixined(RedactMixin):
        _skip_redact = ["nested_field"]
        password: str
        nested_field: Any

    assert redactor.redact(Mixined()).password == "passwd"
    assert redactor.redact(Mixined()).transparent == "123"
    assert redactor.redact({"password": "passwd"}) == {"password": "***"}
    redacted_nested = redactor.redact(NestedMixined("passwd", NestedMixined("passwd", None)))
    assert redacted_nested == NestedMixined("***", NestedMixined("passwd", None))


def test_get_fully_qualified_class_name():
    from airflow.providers.openlineage.plugins.adapter import OpenLineageAdapter

    result = get_fully_qualified_class_name(BashOperator(task_id="test", bash_command="exit 0;"))
    assert result == "airflow.providers.standard.operators.bash.BashOperator"

    result = get_fully_qualified_class_name(OpenLineageAdapter())
    assert result == "airflow.providers.openlineage.plugins.adapter.OpenLineageAdapter"


@patch("airflow.providers.openlineage.conf.disabled_operators")
def test_is_operator_disabled(mock_disabled_operators):
    mock_disabled_operators.return_value = {}
    op = BashOperator(task_id="test", bash_command="exit 0;")
    assert is_operator_disabled(op) is False

    mock_disabled_operators.return_value = {"random_string"}
    assert is_operator_disabled(op) is False

    mock_disabled_operators.return_value = {
        "airflow.providers.standard.operators.bash.BashOperator",
        "airflow.providers.standard.operators.python.PythonOperator",
    }
    assert is_operator_disabled(op) is True


@patch("airflow.providers.openlineage.conf.include_full_task_info")
def test_includes_full_task_info(mock_include_full_task_info):
    mock_include_full_task_info.return_value = True
    # There should be no 'bash_command' in excludes and it's not in includes - so
    # it's a good choice for checking TaskInfo vs TaskInfoComplete
    assert (
        "bash_command"
        in get_airflow_run_facet(
            MagicMock(),
            MagicMock(),
            MagicMock(),
            BashOperator(task_id="bash_op", bash_command="sleep 1"),
            MagicMock(),
        )["airflow"].task
    )


@patch("airflow.providers.openlineage.conf.include_full_task_info")
def test_does_not_include_full_task_info(mock_include_full_task_info):
    mock_include_full_task_info.return_value = False
    # There should be no 'bash_command' in excludes and it's not in includes - so
    # it's a good choice for checking TaskInfo vs TaskInfoComplete
    assert (
        "bash_command"
        not in get_airflow_run_facet(
            MagicMock(),
            MagicMock(),
            MagicMock(),
            BashOperator(task_id="bash_op", bash_command="sleep 1"),
            MagicMock(),
        )["airflow"].task
    )


@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="This test checks serialization only in 3.0 conditions")
def test_serialize_timetable_complex_with_alias():
    from airflow.providers.common.compat.assets import AssetAlias, AssetAll, AssetAny
    from airflow.timetables.simple import AssetTriggeredTimetable

    asset = AssetAny(
        Asset(name="2", uri="test://2", group="test-group"),
        AssetAlias(name="example-alias", group="test-group"),
        Asset(name="3", uri="test://3", group="test-group"),
        AssetAll(AssetAlias("another"), Asset("4")),
    )
    dag = MagicMock()
    dag.timetable = AssetTriggeredTimetable(asset)
    dag_info = DagInfo(dag)

    assert dag_info.timetable == {
        "asset_condition": {
            "__type": DagAttributeTypes.ASSET_ANY,
            "objects": [
                {
                    "__type": DagAttributeTypes.ASSET,
                    "extra": {},
                    "uri": "test://2/",
                    "name": "2",
                    "group": "test-group",
                },
                {
                    "__type": DagAttributeTypes.ASSET_ALIAS,
                    "name": "example-alias",
                    "group": "test-group",
                },
                {
                    "__type": DagAttributeTypes.ASSET,
                    "extra": {},
                    "uri": "test://3/",
                    "name": "3",
                    "group": "test-group",
                },
                {
                    "__type": DagAttributeTypes.ASSET_ALL,
                    "objects": [
                        {
                            "__type": DagAttributeTypes.ASSET_ALIAS,
                            "name": "another",
                            "group": "asset",
                        },
                        {
                            "__type": DagAttributeTypes.ASSET,
                            "extra": {},
                            "uri": "4",
                            "name": "4",
                            "group": "asset",
                        },
                    ],
                },
            ],
        }
    }


@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="This test checks serialization only in 3.0 conditions")
def test_serialize_timetable_single_asset():
    dag = DAG(dag_id="test", start_date=datetime.datetime(2025, 1, 1), schedule=Asset("a"))
    dag_info = DagInfo(dag)
    assert dag_info.timetable == {
        "asset_condition": {
            "__type": DagAttributeTypes.ASSET,
            "uri": "a",
            "name": "a",
            "group": "asset",
            "extra": {},
        }
    }


@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="This test checks serialization only in 3.0 conditions")
def test_serialize_timetable_list_of_assets():
    dag = DAG(dag_id="test", start_date=datetime.datetime(2025, 1, 1), schedule=[Asset("a"), Asset("b")])
    dag_info = DagInfo(dag)
    assert dag_info.timetable == {
        "asset_condition": {
            "__type": DagAttributeTypes.ASSET_ALL,
            "objects": [
                {"__type": DagAttributeTypes.ASSET, "uri": "a", "name": "a", "group": "asset", "extra": {}},
                {"__type": DagAttributeTypes.ASSET, "uri": "b", "name": "b", "group": "asset", "extra": {}},
            ],
        }
    }


@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="This test checks serialization only in 3.0 conditions")
def test_serialize_timetable_with_complex_logical_condition():
    dag = DAG(
        dag_id="test",
        start_date=datetime.datetime(2025, 1, 1),
        schedule=(Asset("ds1", extra={"some_extra": 1}) | Asset("ds2"))
        & (Asset("ds3") | Asset("ds4", extra={"another_extra": 345})),
    )
    dag_info = DagInfo(dag)
    assert dag_info.timetable == {
        "asset_condition": {
            "__type": DagAttributeTypes.ASSET_ALL,
            "objects": [
                {
                    "__type": DagAttributeTypes.ASSET_ANY,
                    "objects": [
                        {
                            "__type": DagAttributeTypes.ASSET,
                            "uri": "ds1",
                            "extra": {"some_extra": 1},
                            "name": "ds1",
                            "group": "asset",
                        },
                        {
                            "__type": DagAttributeTypes.ASSET,
                            "uri": "ds2",
                            "extra": {},
                            "name": "ds2",
                            "group": "asset",
                        },
                    ],
                },
                {
                    "__type": DagAttributeTypes.ASSET_ANY,
                    "objects": [
                        {
                            "__type": DagAttributeTypes.ASSET,
                            "uri": "ds3",
                            "extra": {},
                            "name": "ds3",
                            "group": "asset",
                        },
                        {
                            "__type": DagAttributeTypes.ASSET,
                            "uri": "ds4",
                            "extra": {"another_extra": 345},
                            "name": "ds4",
                            "group": "asset",
                        },
                    ],
                },
            ],
        }
    }


@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="This test checks serialization only in 3.0 conditions")
def test_serialize_timetable_with_dataset_or_time_schedule():
    from airflow.timetables.assets import AssetOrTimeSchedule
    from airflow.timetables.trigger import CronTriggerTimetable

    dag = DAG(
        dag_id="test",
        start_date=datetime.datetime(2025, 1, 1),
        schedule=AssetOrTimeSchedule(
            timetable=CronTriggerTimetable("0 0 * 3 *", timezone="UTC"),
            assets=(Asset("ds1", extra={"some_extra": 1}) | Asset("ds2"))
            & (Asset("ds3") | Asset("ds4", extra={"another_extra": 345})),
        ),
    )
    dag_info = DagInfo(dag)
    assert dag_info.timetable == {
        "timetable": {
            Encoding.TYPE: "airflow.timetables.trigger.CronTriggerTimetable",
            Encoding.VAR: {
                "expression": "0 0 * 3 *",
                "timezone": "UTC",
                "interval": 0.0,
                "run_immediately": False,
            },
        },
        "asset_condition": {
            "__type": DagAttributeTypes.ASSET_ALL,
            "objects": [
                {
                    "__type": DagAttributeTypes.ASSET_ANY,
                    "objects": [
                        {
                            "__type": DagAttributeTypes.ASSET,
                            "uri": "ds1",
                            "extra": {"some_extra": 1},
                            "name": "ds1",
                            "group": "asset",
                        },
                        {
                            "__type": DagAttributeTypes.ASSET,
                            "uri": "ds2",
                            "extra": {},
                            "name": "ds2",
                            "group": "asset",
                        },
                    ],
                },
                {
                    "__type": DagAttributeTypes.ASSET_ANY,
                    "objects": [
                        {
                            "__type": DagAttributeTypes.ASSET,
                            "uri": "ds3",
                            "extra": {},
                            "name": "ds3",
                            "group": "asset",
                        },
                        {
                            "__type": DagAttributeTypes.ASSET,
                            "uri": "ds4",
                            "extra": {"another_extra": 345},
                            "name": "ds4",
                            "group": "asset",
                        },
                    ],
                },
            ],
        },
    }


@pytest.mark.skipif(
    AIRFLOW_V_3_0_PLUS,
    reason="This test checks serialization only in 2.10 conditions",
)
def test_serialize_timetable_2_10_complex_with_alias():
    from airflow.providers.common.compat.assets import AssetAlias, AssetAll, AssetAny
    from airflow.timetables.simple import DatasetTriggeredTimetable

    asset = AssetAny(
        Asset("2"),
        AssetAlias("example-alias"),
        Asset("3"),
        AssetAll(AssetAlias("this-should-not-be-seen"), Asset("4")),
    )

    dag = MagicMock()
    dag.timetable = DatasetTriggeredTimetable(asset)
    dag_info = DagInfo(dag)

    assert dag_info.timetable == {
        "dataset_condition": {
            "__type": DagAttributeTypes.DATASET_ANY,
            "objects": [
                {"__type": DagAttributeTypes.DATASET, "extra": None, "uri": "2"},
                {"__type": DagAttributeTypes.DATASET_ANY, "objects": []},
                {"__type": DagAttributeTypes.DATASET, "extra": None, "uri": "3"},
                {
                    "__type": DagAttributeTypes.DATASET_ALL,
                    "objects": [
                        {"__type": DagAttributeTypes.DATASET_ANY, "objects": []},
                        {"__type": DagAttributeTypes.DATASET, "extra": None, "uri": "4"},
                    ],
                },
            ],
        }
    }


@pytest.mark.skipif(
    AIRFLOW_V_3_0_PLUS,
    reason="This test checks serialization only in 2.10 conditions",
)
def test_serialize_timetable_2_10_single_asset():
    dag = DAG(dag_id="test", start_date=datetime.datetime(2025, 1, 1), schedule=Asset("a"))
    dag_info = DagInfo(dag)
    assert dag_info.timetable == {
        "dataset_condition": {"__type": DagAttributeTypes.DATASET, "uri": "a", "extra": None}
    }


@pytest.mark.skipif(
    AIRFLOW_V_3_0_PLUS,
    reason="This test checks serialization only in 2.10 conditions",
)
def test_serialize_timetable_2_10_list_of_assets():
    dag = DAG(dag_id="test", start_date=datetime.datetime(2025, 1, 1), schedule=[Asset("a"), Asset("b")])
    dag_info = DagInfo(dag)
    assert dag_info.timetable == {
        "dataset_condition": {
            "__type": DagAttributeTypes.DATASET_ALL,
            "objects": [
                {"__type": DagAttributeTypes.DATASET, "extra": None, "uri": "a"},
                {"__type": DagAttributeTypes.DATASET, "extra": None, "uri": "b"},
            ],
        }
    }


@pytest.mark.skipif(
    AIRFLOW_V_3_0_PLUS,
    reason="This test checks serialization only in 2.10 conditions",
)
def test_serialize_timetable_2_10_with_complex_logical_condition():
    dag = DAG(
        dag_id="test",
        start_date=datetime.datetime(2025, 1, 1),
        schedule=(Asset("ds1", extra={"some_extra": 1}) | Asset("ds2"))
        & (Asset("ds3") | Asset("ds4", extra={"another_extra": 345})),
    )
    dag_info = DagInfo(dag)
    assert dag_info.timetable == {
        "dataset_condition": {
            "__type": DagAttributeTypes.DATASET_ALL,
            "objects": [
                {
                    "__type": DagAttributeTypes.DATASET_ANY,
                    "objects": [
                        {"__type": DagAttributeTypes.DATASET, "uri": "ds1", "extra": {"some_extra": 1}},
                        {"__type": DagAttributeTypes.DATASET, "uri": "ds2", "extra": None},
                    ],
                },
                {
                    "__type": DagAttributeTypes.DATASET_ANY,
                    "objects": [
                        {"__type": DagAttributeTypes.DATASET, "uri": "ds3", "extra": None},
                        {"__type": DagAttributeTypes.DATASET, "uri": "ds4", "extra": {"another_extra": 345}},
                    ],
                },
            ],
        }
    }


@pytest.mark.skipif(
    AIRFLOW_V_3_0_PLUS,
    reason="This test checks serialization only in 2.10 conditions",
)
def test_serialize_timetable_2_10_with_dataset_or_time_schedule():
    from airflow.timetables.datasets import DatasetOrTimeSchedule
    from airflow.timetables.trigger import CronTriggerTimetable

    dag = DAG(
        dag_id="test",
        start_date=datetime.datetime(2025, 1, 1),
        schedule=DatasetOrTimeSchedule(
            timetable=CronTriggerTimetable("0 0 * 3 *", timezone="UTC"),
            datasets=(Asset("ds1", extra={"some_extra": 1}) | Asset("ds2"))
            & (Asset("ds3") | Asset("ds4", extra={"another_extra": 345})),
        ),
    )
    dag_info = DagInfo(dag)
    assert dag_info.timetable == {
        "timetable": {
            "__type": "airflow.timetables.trigger.CronTriggerTimetable",
            "__var": {"expression": "0 0 * 3 *", "timezone": "UTC", "interval": 0.0},
        },
        "dataset_condition": {
            "__type": DagAttributeTypes.DATASET_ALL,
            "objects": [
                {
                    "__type": DagAttributeTypes.DATASET_ANY,
                    "objects": [
                        {"__type": DagAttributeTypes.DATASET, "uri": "ds1", "extra": {"some_extra": 1}},
                        {"__type": DagAttributeTypes.DATASET, "uri": "ds2", "extra": None},
                    ],
                },
                {
                    "__type": DagAttributeTypes.DATASET_ANY,
                    "objects": [
                        {"__type": DagAttributeTypes.DATASET, "uri": "ds3", "extra": None},
                        {"__type": DagAttributeTypes.DATASET, "uri": "ds4", "extra": {"another_extra": 345}},
                    ],
                },
            ],
        },
    }


@pytest.mark.parametrize(
    ("airflow_version", "ol_version"),
    [
        ("2.9.3", "1.12.2"),
        ("2.10.1", "1.13.0"),
        ("3.0.0", "1.14.0"),
    ],
)
def test_get_processing_engine_facet(airflow_version, ol_version):
    with patch("airflow.providers.openlineage.utils.utils.AIRFLOW_VERSION", airflow_version):
        with patch("airflow.providers.openlineage.utils.utils.OPENLINEAGE_PROVIDER_VERSION", ol_version):
            result = get_processing_engine_facet()
            assert result["processing_engine"].version == airflow_version
            assert result["processing_engine"].openlineageAdapterVersion == ol_version



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

import attrs

from airflow.providers.common.compat.openlineage.facet import RunFacet

if TYPE_CHECKING:
    from airflow.models.taskinstance import TaskInstance, TaskInstanceState


@attrs.define
class MyCustomRunFacet(RunFacet):
    """Define a custom run facet."""

    name: str
    cluster: str


def get_additional_test_facet(
    task_instance: TaskInstance, ti_state: TaskInstanceState
) -> dict[str, RunFacet] | None:
    operator_name = task_instance.task.operator_name if task_instance.task else ""
    if operator_name == "BashOperator":
        return None
    return {
        "additional_run_facet": MyCustomRunFacet(
            name=f"test-lineage-namespace-{ti_state}",
            cluster=f"TEST_{task_instance.dag_id}.{task_instance.task_id}",
        )
    }


def get_duplicate_test_facet_key(
    task_instance: TaskInstance, ti_state: TaskInstanceState
) -> dict[str, RunFacet] | None:
    return get_additional_test_facet(task_instance, ti_state)


def get_another_test_facet(task_instance, ti_state):
    return {"another_run_facet": {"name": "another-lineage-namespace"}}


def return_type_is_not_dict(task_instance, ti_state):
    return "return type is not dict"


def get_custom_facet_throws_exception(task_instance, ti_state):
    raise Exception("fake exception from custom facet function")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Simple DAG with deferrable operator.

It checks:
    - that at least two task START events (before and after deferral) are emitted and
      the try_num remains at 1
"""

from __future__ import annotations

import warnings
from datetime import datetime, timedelta

from airflow import DAG
from airflow.exceptions import AirflowProviderDeprecationWarning
from airflow.models import Variable
from airflow.providers.standard.operators.python import PythonOperator
from airflow.providers.standard.sensors.time_delta import TimeDeltaSensorAsync

from system.openlineage.expected_events import get_expected_event_file_path
from system.openlineage.operator import OpenLineageTestOperator


def check_events_number_func():
    events = Variable.get(key="openlineage_defer_simple_dag.wait.event.start", deserialize_json=True)
    if len(events) < 2:
        raise ValueError(f"Expected at least 2 START events for task `wait`, got {len(events)}")


DAG_ID = "openlineage_defer_simple_dag"

with DAG(
    dag_id=DAG_ID,
    start_date=datetime(2021, 1, 1),
    schedule=None,
    catchup=False,
    default_args={"retries": 0},
) as dag:
    # Timedelta is compared to the DAGRun start timestamp, which can occur long before a worker picks up the
    # task. We need to ensure the sensor gets deferred at least once, so setting 180s.
    with warnings.catch_warnings():  # TODO Switch to TimeDeltaSensor when deferrable is released
        warnings.simplefilter("ignore", AirflowProviderDeprecationWarning)
        wait = TimeDeltaSensorAsync(task_id="wait", delta=timedelta(seconds=180))

    check_events_number = PythonOperator(
        task_id="check_events_number", python_callable=check_events_number_func
    )

    check_events = OpenLineageTestOperator(
        task_id="check_events",
        file_path=get_expected_event_file_path(DAG_ID),
        allow_duplicate_events=True,
    )

    wait >> check_events_number >> check_events


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
import time

from airflow.models.dag import DAG
from airflow.models.operator import BaseOperator
from airflow.providers.common.compat.openlineage.facet import Dataset
from airflow.providers.openlineage.extractors import OperatorLineage


class OpenLineageExecutionOperator(BaseOperator):
    def __init__(self, *, stall_amount=0, fail=False, **kwargs) -> None:
        super().__init__(**kwargs)
        self.stall_amount = stall_amount
        self.fail = fail

    def execute(self, context):
        self.log.error("STALL AMOUNT %s", self.stall_amount)
        time.sleep(1)
        if self.fail:
            raise Exception("Failed")

    def get_openlineage_facets_on_start(self):
        return OperatorLineage(inputs=[Dataset(namespace="test", name="on-start")])

    def get_openlineage_facets_on_complete(self, task_instance):
        self.log.error("STALL AMOUNT %s", self.stall_amount)
        time.sleep(self.stall_amount)
        return OperatorLineage(inputs=[Dataset(namespace="test", name="on-complete")])

    def get_openlineage_facets_on_failure(self, task_instance):
        self.log.error("STALL AMOUNT %s", self.stall_amount)
        time.sleep(self.stall_amount)
        return OperatorLineage(inputs=[Dataset(namespace="test", name="on-failure")])


with DAG(
    dag_id="test_openlineage_execution",
    default_args={"owner": "airflow", "retries": 3, "start_date": datetime.datetime(2022, 1, 1)},
    schedule="0 0 * * *",
    dagrun_timeout=datetime.timedelta(minutes=60),
):
    no_stall = OpenLineageExecutionOperator(task_id="execute_no_stall")

    short_stall = OpenLineageExecutionOperator(task_id="execute_short_stall", stall_amount=5)

    mid_stall = OpenLineageExecutionOperator(task_id="execute_mid_stall", stall_amount=15)

    long_stall = OpenLineageExecutionOperator(task_id="execute_long_stall", stall_amount=30)

    fail = OpenLineageExecutionOperator(task_id="execute_fail", fail=True)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import tempfile
from typing import TYPE_CHECKING, Any
from unittest.mock import MagicMock

import pytest
from openlineage.client.event_v2 import Dataset as OpenLineageDataset
from openlineage.client.facet_v2 import (
    documentation_dataset,
    ownership_dataset,
    schema_dataset,
)

from airflow.models.taskinstance import TaskInstance
from airflow.providers.common.compat.lineage.entities import Column, File, Table, User
from airflow.providers.openlineage.extractors import OperatorLineage
from airflow.providers.openlineage.extractors.manager import ExtractorManager
from airflow.providers.openlineage.utils.utils import Asset
from airflow.utils.state import State

from tests_common.test_utils.compat import DateTimeSensor, PythonOperator
from tests_common.test_utils.markers import skip_if_force_lowest_dependencies_marker
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if TYPE_CHECKING:
    try:
        from airflow.sdk.api.datamodels._generated import AssetEventDagRunReference, TIRunContext
        from airflow.sdk.definitions.context import Context

    except ImportError:
        # TODO: Remove once provider drops support for Airflow 2
        # TIRunContext is only used in Airflow 3 tests
        from airflow.utils.context import Context

        AssetEventDagRunReference = TIRunContext = Any  # type: ignore[misc, assignment]


@pytest.fixture
def hook_lineage_collector():
    from airflow.lineage import hook
    from airflow.providers.common.compat.lineage.hook import (
        get_hook_lineage_collector,
    )

    hook._hook_lineage_collector = None
    hook._hook_lineage_collector = hook.HookLineageCollector()

    yield get_hook_lineage_collector()

    hook._hook_lineage_collector = None


if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import BaseOperator, ObjectStoragePath
    from airflow.sdk.api.datamodels._generated import TaskInstance as SDKTaskInstance
    from airflow.sdk.execution_time import task_runner
    from airflow.sdk.execution_time.comms import StartupDetails
    from airflow.sdk.execution_time.task_runner import RuntimeTaskInstance, parse
else:
    from airflow.io.path import ObjectStoragePath  # type: ignore[no-redef]
    from airflow.models import BaseOperator

    SDKTaskInstance = ...  # type: ignore
    task_runner = ...  # type: ignore
    StartupDetails = ...  # type: ignore
    RuntimeTaskInstance = ...  # type: ignore
    parse = ...  # type: ignore


@pytest.mark.parametrize(
    ("uri", "dataset"),
    (
        (
            "s3://bucket1/dir1/file1",
            OpenLineageDataset(namespace="s3://bucket1", name="dir1/file1"),
        ),
        (
            "gs://bucket2/dir2/file2",
            OpenLineageDataset(namespace="gs://bucket2", name="dir2/file2"),
        ),
        (
            "gcs://bucket3/dir3/file3",
            OpenLineageDataset(namespace="gs://bucket3", name="dir3/file3"),
        ),
        (
            "hdfs://namenodehost:8020/file1",
            OpenLineageDataset(namespace="hdfs://namenodehost:8020", name="file1"),
        ),
        (
            "hdfs://namenodehost/file2",
            OpenLineageDataset(namespace="hdfs://namenodehost", name="file2"),
        ),
        (
            "file://localhost/etc/fstab",
            OpenLineageDataset(namespace="file://localhost", name="etc/fstab"),
        ),
        (
            "file:///etc/fstab",
            OpenLineageDataset(namespace="file://", name="etc/fstab"),
        ),
        ("https://test.com", OpenLineageDataset(namespace="https", name="test.com")),
        (
            "https://test.com?param1=test1&param2=test2",
            OpenLineageDataset(namespace="https", name="test.com"),
        ),
        ("file:test.csv", None),
        ("not_an_url", None),
    ),
)
def test_convert_to_ol_dataset_from_object_storage_uri(uri, dataset):
    result = ExtractorManager.convert_to_ol_dataset_from_object_storage_uri(uri)
    assert result == dataset


@pytest.mark.parametrize(
    ("obj", "dataset"),
    (
        (
            OpenLineageDataset(namespace="n1", name="f1"),
            OpenLineageDataset(namespace="n1", name="f1"),
        ),
        (
            File(url="s3://bucket1/dir1/file1"),
            OpenLineageDataset(namespace="s3://bucket1", name="dir1/file1"),
        ),
        (
            File(url="gs://bucket2/dir2/file2"),
            OpenLineageDataset(namespace="gs://bucket2", name="dir2/file2"),
        ),
        (
            File(url="gcs://bucket3/dir3/file3"),
            OpenLineageDataset(namespace="gs://bucket3", name="dir3/file3"),
        ),
        (
            File(url="hdfs://namenodehost:8020/file1"),
            OpenLineageDataset(namespace="hdfs://namenodehost:8020", name="file1"),
        ),
        (
            File(url="hdfs://namenodehost/file2"),
            OpenLineageDataset(namespace="hdfs://namenodehost", name="file2"),
        ),
        (
            File(url="file://localhost/etc/fstab"),
            OpenLineageDataset(namespace="file://localhost", name="etc/fstab"),
        ),
        (
            File(url="file:///etc/fstab"),
            OpenLineageDataset(namespace="file://", name="etc/fstab"),
        ),
        (
            File(url="https://test.com"),
            OpenLineageDataset(namespace="https", name="test.com"),
        ),
        (
            Table(cluster="c1", database="d1", name="t1"),
            OpenLineageDataset(namespace="c1", name="d1.t1"),
        ),
        ("gs://bucket2/dir2/file2", None),
        ("not_an_url", None),
    ),
)
def test_convert_to_ol_dataset(obj, dataset):
    result = ExtractorManager.convert_to_ol_dataset(obj)
    assert result == dataset


def test_convert_to_ol_dataset_from_table_with_columns_and_owners():
    table = Table(
        cluster="c1",
        database="d1",
        name="t1",
        columns=[
            Column(name="col1", description="desc1", data_type="type1"),
            Column(name="col2", description="desc2", data_type="type2"),
        ],
        owners=[
            User(email="mike@company.com", first_name="Mike", last_name="Smith"),
            User(email="theo@company.com", first_name="Theo"),
            User(email="smith@company.com", last_name="Smith"),
            User(email="jane@company.com"),
        ],
        description="test description",
    )
    expected_facets = {
        "schema": schema_dataset.SchemaDatasetFacet(
            fields=[
                schema_dataset.SchemaDatasetFacetFields(
                    name="col1",
                    type="type1",
                    description="desc1",
                ),
                schema_dataset.SchemaDatasetFacetFields(
                    name="col2",
                    type="type2",
                    description="desc2",
                ),
            ]
        ),
        "ownership": ownership_dataset.OwnershipDatasetFacet(
            owners=[
                ownership_dataset.Owner(name="user:Mike Smith <mike@company.com>", type=""),
                ownership_dataset.Owner(name="user:Theo <theo@company.com>", type=""),
                ownership_dataset.Owner(name="user:Smith <smith@company.com>", type=""),
                ownership_dataset.Owner(name="user:<jane@company.com>", type=""),
            ]
        ),
        "documentation": documentation_dataset.DocumentationDatasetFacet(description="test description"),
    }
    result = ExtractorManager.convert_to_ol_dataset_from_table(table)
    assert result.namespace == "c1"
    assert result.name == "d1.t1"
    assert result.facets == expected_facets


def test_convert_to_ol_dataset_table():
    table = Table(
        cluster="c1",
        database="d1",
        name="t1",
        columns=[
            Column(name="col1", description="desc1", data_type="type1"),
            Column(name="col2", description="desc2", data_type="type2"),
        ],
        owners=[
            User(email="mike@company.com", first_name="Mike", last_name="Smith"),
            User(email="theo@company.com", first_name="Theo"),
            User(email="smith@company.com", last_name="Smith"),
            User(email="jane@company.com"),
        ],
    )
    expected_facets = {
        "schema": schema_dataset.SchemaDatasetFacet(
            fields=[
                schema_dataset.SchemaDatasetFacetFields(
                    name="col1",
                    type="type1",
                    description="desc1",
                ),
                schema_dataset.SchemaDatasetFacetFields(
                    name="col2",
                    type="type2",
                    description="desc2",
                ),
            ]
        ),
        "ownership": ownership_dataset.OwnershipDatasetFacet(
            owners=[
                ownership_dataset.Owner(name="user:Mike Smith <mike@company.com>", type=""),
                ownership_dataset.Owner(name="user:Theo <theo@company.com>", type=""),
                ownership_dataset.Owner(name="user:Smith <smith@company.com>", type=""),
                ownership_dataset.Owner(name="user:<jane@company.com>", type=""),
            ]
        ),
    }

    result = ExtractorManager.convert_to_ol_dataset(table)
    assert result.namespace == "c1"
    assert result.name == "d1.t1"
    assert result.facets == expected_facets


@skip_if_force_lowest_dependencies_marker
def test_extractor_manager_uses_hook_level_lineage(hook_lineage_collector):
    dagrun = MagicMock()
    task = MagicMock()
    del task.get_openlineage_facets_on_start
    del task.get_openlineage_facets_on_complete
    ti = MagicMock()

    hook_lineage_collector.add_input_asset(None, uri="s3://bucket/input_key")
    hook_lineage_collector.add_output_asset(None, uri="s3://bucket/output_key")
    extractor_manager = ExtractorManager()
    metadata = extractor_manager.extract_metadata(
        dagrun=dagrun, task=task, task_instance_state=None, task_instance=ti
    )

    assert metadata.inputs == [OpenLineageDataset(namespace="s3://bucket", name="input_key")]
    assert metadata.outputs == [OpenLineageDataset(namespace="s3://bucket", name="output_key")]


def test_extractor_manager_does_not_use_hook_level_lineage_when_operator(
    hook_lineage_collector,
):
    class FakeSupportedOperator(BaseOperator):
        def execute(self, context: Context) -> Any:
            pass

        def get_openlineage_facets_on_start(self):
            return OperatorLineage(
                inputs=[OpenLineageDataset(namespace="s3://bucket", name="proper_input_key")]
            )

    dagrun = MagicMock()
    task = FakeSupportedOperator(task_id="test_task_extractor")
    ti = MagicMock()
    hook_lineage_collector.add_input_asset(None, uri="s3://bucket/input_key")

    extractor_manager = ExtractorManager()
    metadata = extractor_manager.extract_metadata(
        dagrun=dagrun, task=task, task_instance_state=None, task_instance=ti
    )

    # s3://bucket/input_key not here - use data from operator
    assert metadata.inputs == [OpenLineageDataset(namespace="s3://bucket", name="proper_input_key")]
    assert metadata.outputs == []


@pytest.mark.skipif(
    AIRFLOW_V_3_0_PLUS,
    reason="Test for hook level lineage in Airflow < 3.0",
)
def test_extractor_manager_gets_data_from_pythonoperator(session, dag_maker, hook_lineage_collector):
    path = None
    with tempfile.NamedTemporaryFile() as f:
        path = f.name
        with dag_maker():

            def use_read():
                storage_path = ObjectStoragePath(path)
                with storage_path.open("w") as out:
                    out.write("test")

            task = PythonOperator(task_id="test_task_extractor_pythonoperator", python_callable=use_read)

    dr = dag_maker.create_dagrun()
    ti = TaskInstance(task=task, run_id=dr.run_id)
    ti.refresh_from_db()
    ti.state = State.QUEUED
    session.merge(ti)
    session.commit()

    ti.run()

    datasets = hook_lineage_collector.collected_assets

    assert len(datasets.outputs) == 1
    assert datasets.outputs[0].asset == Asset(uri=path)


@pytest.mark.db_test
@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Task SDK related test")
def test_extractor_manager_gets_data_from_pythonoperator_tasksdk(session, hook_lineage_collector, run_task):
    path = None
    with tempfile.NamedTemporaryFile() as f:
        path = f.name

        def use_read():
            storage_path = ObjectStoragePath(path)
            with storage_path.open("w") as out:
                out.write("test")

    task = PythonOperator(task_id="test_task_extractor_pythonoperator", python_callable=use_read)

    run_task(task, dag_id="test_hookcollector_dag")

    datasets = hook_lineage_collector.collected_assets

    assert len(datasets.outputs) == 1
    assert datasets.outputs[0].asset == Asset(uri=path)


def test_extract_inlets_and_outlets_with_operator():
    inlets = [OpenLineageDataset(namespace="namespace1", name="name1")]
    outlets = [OpenLineageDataset(namespace="namespace2", name="name2")]

    extractor_manager = ExtractorManager()
    task = PythonOperator(task_id="task_id", python_callable=lambda x: x, inlets=inlets, outlets=outlets)
    lineage = OperatorLineage()
    extractor_manager.extract_inlets_and_outlets(lineage, task)
    assert lineage.inputs == inlets
    assert lineage.outputs == outlets


def test_extract_inlets_and_outlets_with_sensor():
    inlets = [OpenLineageDataset(namespace="namespace1", name="name1")]
    outlets = [OpenLineageDataset(namespace="namespace2", name="name2")]

    extractor_manager = ExtractorManager()
    task = DateTimeSensor(
        task_id="task_id", target_time="2025-04-04T08:48:13.713922+00:00", inlets=inlets, outlets=outlets
    )
    lineage = OperatorLineage()
    extractor_manager.extract_inlets_and_outlets(lineage, task)
    assert lineage.inputs == inlets
    assert lineage.outputs == outlets



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Simple DAG with multiple asset logical condition in list schedule for Airflow 3.
Use of list will result in the whole condition being wrapped with additional `asset_all`.

It checks:
    - schedule serialization
"""

from __future__ import annotations

from datetime import datetime

from airflow import DAG
from airflow.providers.common.compat.assets import Asset
from airflow.providers.standard.operators.bash import BashOperator

from system.openlineage.expected_events import AIRFLOW_VERSION, get_expected_event_file_path
from system.openlineage.operator import OpenLineageTestOperator

if AIRFLOW_VERSION.major == 3:
    schedule = [
        (Asset(uri="s3://bucket/file.txt", extra={"a": 1}) | Asset(uri="s3://bucket2/file.txt"))
        & (Asset(uri="s3://bucket3/file.txt") | Asset(uri="s3://bucket4/file.txt", extra={"b": 2}))
    ]
else:
    # Logical Asset condition wrapped in list breaks DAG processing in Airflow 2 - check is skipped
    schedule = None  # type: ignore[assignment]


DAG_ID = "openlineage_schedule_list_complex_assets_dag"

with DAG(
    dag_id=DAG_ID,
    start_date=datetime(2021, 1, 1),
    schedule=schedule,
    catchup=False,
    default_args={"retries": 0},
) as dag:
    do_nothing_task = BashOperator(task_id="do_nothing_task", bash_command="sleep 10;")

    check_events = OpenLineageTestOperator(
        task_id="check_events", file_path=get_expected_event_file_path(DAG_ID)
    )

    do_nothing_task >> check_events


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime, timezone
from unittest import mock

import pytest

from airflow import __version__
from airflow.providers.openlineage.conf import namespace
from airflow.providers.openlineage.plugins.macros import (
    lineage_job_name,
    lineage_job_namespace,
    lineage_parent_id,
    lineage_root_run_id,
    lineage_run_id,
)

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

_DAG_NAMESPACE = namespace()

if __version__.startswith("2."):
    LOGICAL_DATE_KEY = "execution_date"
else:
    LOGICAL_DATE_KEY = "logical_date"


def test_lineage_job_namespace():
    assert lineage_job_namespace() == _DAG_NAMESPACE


def test_lineage_job_name():
    task_instance = mock.MagicMock(
        dag_id="dag_id",
        task_id="task_id",
        try_number=1,
        **{LOGICAL_DATE_KEY: datetime(2020, 1, 1, 1, 1, 1, 0, tzinfo=timezone.utc)},
    )
    assert lineage_job_name(task_instance) == "dag_id.task_id"


def test_lineage_run_id():
    date = datetime(2020, 1, 1, 1, 1, 1, 0, tzinfo=timezone.utc)
    dag_run = mock.MagicMock(run_id="run_id")
    dag_run.logical_date = date
    task_instance = mock.MagicMock(
        dag_id="dag_id",
        task_id="task_id",
        dag_run=dag_run,
        logical_date=date,
        try_number=1,
    )

    call_result1 = lineage_run_id(task_instance)
    call_result2 = lineage_run_id(task_instance)

    # random part value does not matter, it just has to be the same for the same TaskInstance
    assert call_result1 == call_result2
    # execution_date is used as most significant bits of UUID
    assert call_result1.startswith("016f5e9e-c4c8-")


@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Test only for Airflow 3.0+")
def test_lineage_run_after_airflow_3():
    dag_run = mock.MagicMock(run_id="run_id")
    dag_run.run_after = datetime(2020, 1, 1, 1, 1, 1, 0, tzinfo=timezone.utc)
    dag_run.logical_date = None
    task_instance = mock.MagicMock(
        dag_id="dag_id",
        task_id="task_id",
        dag_run=dag_run,
        try_number=1,
    )

    call_result1 = lineage_run_id(task_instance)
    call_result2 = lineage_run_id(task_instance)

    # random part value does not matter, it just has to be the same for the same TaskInstance
    assert call_result1 == call_result2
    # execution_date is used as most significant bits of UUID
    assert call_result1.startswith("016f5e9e-c4c8-")


@mock.patch("airflow.providers.openlineage.plugins.macros.lineage_run_id")
def test_lineage_parent_id(mock_run_id):
    mock_run_id.return_value = "run_id"
    task_instance = mock.MagicMock(
        dag_id="dag_id",
        task_id="task_id",
        try_number=1,
        **{LOGICAL_DATE_KEY: datetime(2020, 1, 1, 1, 1, 1, 0, tzinfo=timezone.utc)},
    )
    actual = lineage_parent_id(task_instance)
    expected = f"{_DAG_NAMESPACE}/dag_id.task_id/run_id"
    assert actual == expected


@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Test only for Airflow 3.0+")
def test_lineage_root_run_id_with_runtime_task_instance(create_runtime_ti):
    """Test lineage_root_run_id with real RuntimeTaskInstance object doesn't throw AttributeError."""
    from airflow.sdk import BaseOperator

    task = BaseOperator(task_id="test_task")

    runtime_ti = create_runtime_ti(
        task=task,
        dag_id="test_dag",
        run_id="test_run_id",
    )

    # Explicitly test that the function doesn't throw an AttributeError
    # This was the original issue: AttributeError: 'RuntimeTaskInstance' object has no attribute 'dag_run'
    try:
        assert lineage_root_run_id(runtime_ti) is not None
    except AttributeError as e:
        pytest.fail(f"lineage_root_run_id should not throw AttributeError with RuntimeTaskInstance: {e}")
